"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np


def mat_times_vecvar(mat, vecvar):
    """

    :param mat: a numpy ndarray of shape (m, n).
    :param vecvar: a coniclifts Variable of shape (n,) or (n, 1).

    :return: A_vals, A_rows, A_cols so that the coniclifts Expression
        con = (mat @ vecvar >= 0) would compile to
        A_vals, A_rows, A_cols, np.zeros((m,)), [].
    """
    ids = vecvar.scalar_variable_ids
    A_rows = np.tile(np.arange(mat.shape[0]), reps=mat.shape[1])
    A_cols = np.repeat(ids, mat.shape[0]).tolist()
    A_vals = mat.ravel(order='F').tolist()  # stack columns, then tolist
    return A_vals, A_rows, A_cols


def mat_times_vecvar_plus_mat_times_vecvar(mat1, vecvar1, mat2, vecvar2):
    """

    :param mat1: a numpy ndarray of shape (m, n1).
    :param vecvar1: a coniclifts Variable of shape (n1,) or (n1, 1).
    :param mat2: a numpy ndarray of shape (m, n2)
    :param vecvar2: a coniclifts Variable of shape (n2,) or (n2, 1).

    :return: A_vals, A_rows, A_cols so that the coniclifts Expression
        expr = mat1 @ vecvar1 + mat2 @ vecvar2 would have "expr >= 0"
        compile to A_vals, A_rows, A_cols, np.zeros((m,)), [].
    """
    mat = np.hstack([mat1, mat2])
    ids = [v.id for v in vecvar1.scalar_variables()]
    ids += [v.id for v in vecvar2.scalar_variables()]
    A_rows = np.tile(np.arange(mat.shape[0]), reps=mat.shape[1])
    A_cols = np.repeat(ids, mat.shape[0]).tolist()
    A_vals = mat.ravel(order='F').tolist()  # stack columns, then tolist
    return A_vals, A_rows, A_cols


def mat_times_vecvar_minus_vecvar(mat, vecvar1, vecvar2):
    """

    :param mat: a numpy ndarray of shape (m, n).
    :param vecvar1: a coniclifts Variable of shape (n,) or (n, 1).
    :param vecvar2: a coniclifts Variable of shape (m,) or (m, 1).

    :return: A_vals, A_rows, A_cols so that the coniclifts Expression
        expr = mat @ vecvar1 - vecvar2 would have "expr >= 0"
        compile to A_vals, A_rows, A_cols, np.zeros((m,)), [].
    """
    A_vals, A_rows, A_cols = [], [], []
    num_rows = mat.shape[0]
    vec1_ids = vecvar1.scalar_variable_ids
    vec2_ids = vecvar2.scalar_variable_ids
    for j in range(num_rows):
        A_vals += mat[j, :].tolist()
        A_vals.append(-1)
        A_cols += vec1_ids
        A_cols.append(vec2_ids[j])
        A_rows += [j] * (mat.shape[1] + 1)
    A_rows = np.array(A_rows)
    return A_vals, A_rows, A_cols


def mat_times_vecvar_plus_vec_times_singlevar(mat, vecvar, vec, singlevar):
    """

    :param mat: a numpy ndarray of shape (m, n).
    :param vecvar: a coniclifts Variable of shape (n,) or (n, 1)
    :param vec: a numpy ndarray of shape (m,) or (m, 1).
    :param singlevar: a coniclifts Variable of size 1.
    :return:

    Return A_rows, A_cols, A_vals as if they were generated by a compile() call on
    con = (mat @ vecvar + vec * singlevar >= 0).
    """
    var_ids = vecvar.scalar_variable_ids.copy() + [singlevar.scalar_variables()[0].id]
    ids = np.array(var_ids)
    mat = np.hstack([mat, np.reshape(vec, (-1, 1))])
    A_rows = np.tile(np.arange(mat.shape[0]), reps=mat.shape[1])
    A_cols = np.repeat(ids, mat.shape[0]).tolist()
    A_vals = mat.ravel(order='F').tolist()  # stack columns, then tolist
    return A_vals, A_rows, A_cols


def columns_sum_to_vec(mat, vec):
    A_rows, A_cols, A_vals = [], [], []
    m = mat.shape[0]
    if m != vec.size:
        raise RuntimeError('Incompatible dimensions.')
    b = np.zeros(m,)
    for i in range(m):
        # update cols and data to reflect addition of elements in ith row of mat
        svs = mat[i, :].scalar_variables()
        A_cols += [sv.id for sv in svs]
        A_vals += [1] * len(svs)
        # update cols and data to reflect addition of elements from ith element of vec
        #   ith element of vec is a ScalarExpression!
        id2co = [(a.id, co) for a, co in vec[i].atoms_to_coeffs.items()]
        A_cols += [aid for aid, _ in id2co]
        A_vals += [-co for _, co in id2co]  # we are subtracting, after all.
        # update rows with appropriate number of "i"s.
        A_rows += [i] * (len(svs) + len(id2co))
        # update b
        b[i] -= vec[i].offset
    A_rows = np.array(A_rows)
    return A_vals, A_rows, A_cols, b





"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
from sageopt.symbolic.signomials import __EXPONENT_VECTOR_DECIMAL_POINTS__
from sageopt.symbolic.signomials import Signomial
from sageopt.symbolic.polynomials import Polynomial


__EXPONENT_VECTOR_TOLERANCE__ = 10**-(__EXPONENT_VECTOR_DECIMAL_POINTS__ + 1)


def relative_coeff_vector(s, reference_alpha):
    c = np.zeros(reference_alpha.shape[0])
    common, corr = row_correspondence(s.alpha, reference_alpha)
    c[corr] = s.c[common]
    return c


def row_correspondence(alpha1, alpha2):
    """
    It returns lists "common" and "alpha1_to_alpha2" such that

        alpha1[common, :] == alpha2[alpha1_to_alpha2, :].

    This is useful because it allows us to speak of the "i-th" exponent in a meaningful
    way when dealing with Signomials, without having to adopt a canonical ordering for
    exponent vectors.

    :param alpha1: numpy n-d array.
    :param alpha2: a numpy n-d array.
    :return: a list "alpha1_to_alpha2" such that alpha1 == alpha2[alpha1_to_alpha2, :].
    """
    common = []
    alpha1_to_alpha2 = []
    for i, row in enumerate(alpha1):
        # noinspection PyTypeChecker
        shifted = alpha2 - row
        locs = np.where(np.all(np.abs(shifted) < __EXPONENT_VECTOR_TOLERANCE__, axis=1))
        if len(locs[0]) > 0:
            common.append(i)
            loc = locs[0][0]
            alpha1_to_alpha2.append(loc)
    return common, alpha1_to_alpha2


def moment_reduction_array(s_h, h, L):
    """
    :param s_h: a Signomial or Polynomial (likely with coniclifts Variables as coefficients)
    :param h: a Signomial or Polynomial (with only scalars as coefficients)
    :param L: a Signomial or Polynomial (likely with coniclifts Variables as coefficients)

    Assumptions:

        The rows of lagrangian.alpha must include all rows of (s_h * h).alpha.
        s_h, h, and lagrangian are all of the same type (i.e. all Signomials, or all Polynomials).

    Primary Usage:

        A constraint "h(x) == 0" or "h(x) >= 0" appears in an optimization problem. That constraint
        has been incorporated into a Lagrangian "L", with an associated Lagrange muliplier "s_h".

        Let w(x) = s_h(x) * h(x), and let F be the nonlinear map (consisting entrywise of signomials,
        or polynomials) satisfying w(x) = s_h.c @ F(x). Such a map certainly exists, although its
        entries may include complicated functions with multiple terms (i.e. not just monomials).

        Let G be either G(x) = x^L.alpha (if L is a Polynomial) or G(x) = exp(L.alpha @ x)
        (if L is a Signomial), so that L(x) = L.c @ G(x).

        Since w.alpha is a subset of L.alpha, there exists a matrix C so that F(x) == C @ G(x),
        and in turn so that w(x) == s_h.c @ (C @ G(x)).

        This function returns that matrix C, for use in constructing dual SAGE relaxations.

        The specific usage of C is as follows. Suppose that "v" is the dual variable to the constraint
        that "lagrangian is SAGE". If s_h was a SAGE function, the dual problem will include
        a constraint of the form "C @ v is in the dual SAGE cone over exponents s_h.alpha".
        If s_h was unconstrained, then the dual problem will include a constraint "C @ v == 0".

    Secondary usage:

        We want to minimize the function s_h over some X \\subset \\R^n.
        The function "h" is fixed, and is known to be positive almost-everywhere on X.
        s_h and h are related to L by the identity L == s_h * h.

        Returns a matrix "C" that maps outer-approximations of the moment-cone for functions
        "like" L (i.e. Polynomials or Signomials over exponents L.alpha) to outer-approximations
        of the moment-cone for functions "like" s_h (Polynomials or Signomials over exponents
        s_h.alpha).
    """
    if isinstance(h, Polynomial):
        constructor = Polynomial
    elif isinstance(h, Signomial):
        constructor = Signomial
    else:
        raise RuntimeError('Unknown argument.')
    minimal_L = s_h * h
    for row in minimal_L.alpha_c:
        if row not in L.alpha_c:
            msg0 = 'The product s_h * h contains an exponent vector \n'
            msg1 = '\t ' + str(row) + '\n'
            msg2 = 'that is not present among the exponent vectors of "L".'
            raise RuntimeError(msg0 + msg1 + msg2)
    C_rows = []
    for alpha_i in s_h.alpha_c:
        temp_func = constructor({alpha_i: 1}) * h
        c_row = relative_coeff_vector(temp_func, L.alpha)
        C_rows.append(c_row)
    C = np.vstack(C_rows)
    # Might it be useful to return the matrix of "alpha_i" so that the meaning of C can be understood?
    return C



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from sageopt.coniclifts.constraints.set_membership.setmem import SetMembership
from sageopt.coniclifts.base import Variable, Expression
from sageopt.coniclifts.cones import Cone
from sageopt.coniclifts.operators import affine as aff
from sageopt.coniclifts.operators.precompiled.relent import sum_relent, elementwise_relent
from sageopt.coniclifts.operators.precompiled import affine as compiled_aff
from sageopt.coniclifts.problems.problem import Problem
from sageopt.coniclifts.standards.constants import maximize as CL_MAX, solved as CL_SOLVED
import numpy as np
import scipy.special as special_functions
import warnings


class PrimalSageCone(SetMembership):

    def __init__(self, c, alpha, name, expcovers=None):
        self.name = name
        self.alpha = alpha
        self.m = alpha.shape[0]
        self.n = alpha.shape[1]
        self.c = Expression(c)  # self.c is now definitely an ndarray of ScalarExpressions.
        self.ech = ExpCoverHelper(self.alpha, self.c, expcovers)
        self.nu_vars = dict()
        self.c_vars = dict()
        self.relent_epi_vars = dict()
        self.age_vectors = dict()
        self._variables = self.c.variables()
        self._initialize_variables()
        pass

    def _initialize_variables(self):
        if self.m > 2:
            for i in self.ech.U_I:
                nu_len = np.count_nonzero(self.ech.expcovers[i])
                if nu_len > 0:
                    nu_i = Variable(shape=(nu_len,), name='nu^{(' + str(i) + ')}_{' + self.name + '}')
                    self.nu_vars[i] = nu_i
                    epi_i = Variable(shape=(nu_len,), name='_relent_epi_^{(' + str(i) + ')}_{' + self.name + '}')
                    self.relent_epi_vars[i] = epi_i
                c_len = nu_len
                if i not in self.ech.N_I:
                    c_len += 1
                self.c_vars[i] = Variable(shape=(c_len,), name='c^{(' + str(i) + ')}_{' + self.name + '}')
            self._variables += list(self.nu_vars.values())
            self._variables += list(self.c_vars.values())
            self._variables += list(self.relent_epi_vars.values())
        pass

    def _build_aligned_age_vectors(self):
        for i in self.ech.U_I:
            ci_expr = Expression(np.zeros(self.m,))
            if i in self.ech.N_I:
                ci_expr[self.ech.expcovers[i]] = self.c_vars[i]
                ci_expr[i] = self.c[i]
            else:
                ci_expr[self.ech.expcovers[i]] = self.c_vars[i][:-1]
                ci_expr[i] = self.c_vars[i][-1]
            self.age_vectors[i] = ci_expr
        pass

    def _age_rel_ent_cone_data(self, i):
        idx_set = self.ech.expcovers[i]
        if np.any(idx_set):
            x = self.nu_vars[i]
            y = np.exp(1) * self.age_vectors[i][idx_set]  # This line consumes a large amount of runtime
            z = -self.age_vectors[i][i]
            epi = self.relent_epi_vars[i]
            A_vals, A_rows, A_cols, b, K, epi = sum_relent(x, y, z, epi)
        else:
            # just constrain -age_vectors[i][i] <= 0.
            A_vals, A_rows, A_cols = [1], np.array([0]), [self.age_vectors[i][i].scalar_variables()[0].id]
            b = np.array([0])
            K = [Cone('+', 1)]
        return A_vals, A_rows, A_cols, b, K

    def _age_lin_eq_cone_data(self, i):
        # TODO: use the precompiled operator
        idx_set = self.ech.expcovers[i]
        matrix = (self.alpha[idx_set, :] - self.alpha[i, :]).T
        A_rows = np.tile(np.arange(matrix.shape[0]), reps=matrix.shape[1])
        ids = np.array(self.nu_vars[i].scalar_variable_ids)
        A_cols = np.repeat(ids, matrix.shape[0]).tolist()
        A_vals = matrix.ravel(order='F').tolist()  # stack columns, then tolist
        b = np.zeros(matrix.shape[0], )
        K = [Cone('0', matrix.shape[0])]
        return A_vals, A_rows, A_cols, b, K

    def _age_violation(self, i, norm_ord, c_i):
        if np.any(self.ech.expcovers[i]):
            idx_set = self.ech.expcovers[i]
            x_i = self.nu_vars[i].value
            x_i[x_i < 0] = 0
            y_i = np.exp(1) * c_i[idx_set]
            relent_res = np.sum(special_functions.rel_entr(x_i, y_i)) - c_i[i]  # <= 0
            relent_viol = abs(max(relent_res, 0))
            eq_res = (self.alpha[idx_set, :] - self.alpha[i, :]).T @ x_i  # == 0
            eq_res = eq_res.reshape((-1,))
            eq_viol = np.linalg.norm(eq_res, ord=norm_ord)
            total_viol = relent_viol + eq_viol
            return total_viol
        else:
            c_i = float(self.c_vars[i].value)  # >= 0
            return abs(min(0, c_i))

    def _age_vectors_sum_to_c(self):
        nonconst_locs = np.ones(self.m, dtype=bool)
        nonconst_locs[self.ech.N_I] = False
        aux_c_vars = list(self.age_vectors.values())
        aux_c_vars = aff.vstack(aux_c_vars).T
        aux_c_vars = aux_c_vars[nonconst_locs, :]
        main_c_var = self.c[nonconst_locs]
        A_vals, A_rows, A_cols, b = compiled_aff.columns_sum_to_vec(mat=aux_c_vars, vec=main_c_var)
        K = [Cone('0', b.size)]
        return A_vals, np.array(A_rows), A_cols, b, K

    def variables(self):
        return self._variables

    def conic_form(self):
        if self.m > 2:
            # Lift c_vars and nu_vars into Expressions of length self.m
            self._build_aligned_age_vectors()
            # Record all relative entropy constraints
            cone_data = []
            # age cones
            for i in self.ech.U_I:
                cone_data.append(self._age_rel_ent_cone_data(i))
                if np.any(self.ech.expcovers[i]):
                    cone_data.append(self._age_lin_eq_cone_data(i))
            # Vectors sum to s.c
            cone_data.append(self._age_vectors_sum_to_c())
            return cone_data
        else:
            con = self.c >= 0
            con.epigraph_checked = True
            A_vals, A_rows, A_cols, b, K = con.conic_form()
            cone_data = [(A_vals, A_rows, A_cols, b, K)]
            return cone_data

    def violation(self, norm_ord=np.inf, rough=False):
        c = self.c.value
        if self.m > 2:
            if not rough and c in self:
                return 0
            # compute violation for "AGE vectors sum to c"
            #   Although, we can use the fact that the SAGE cone contains R^m_++.
            #   and so only compute violation for "AGE vectors sum to <= c"
            age_vectors = {i: v.value for i, v in self.age_vectors.items()}
            sum_age_vectors = sum(age_vectors.values())
            residual = c - sum_age_vectors  # want >= 0
            residual[residual < 0] = 0
            sum_to_c_viol = np.linalg.norm(residual, ord=norm_ord)
            # compute violations for each AGE cone
            age_viols = np.zeros(shape=(len(self.ech.U_I,)))
            for idx, i in enumerate(self.ech.U_I):
                age_viols[idx] = self._age_violation(i, norm_ord, age_vectors[i])
            # add the max "AGE violation" to the violation for "AGE vectors sum to c".
            total_viol = sum_to_c_viol + np.max(age_viols)
            return total_viol
        else:
            residual = c.reshape((-1,))  # >= 0
            residual[residual >= 0] = 0
            return np.linalg.norm(c, ord=norm_ord)
        pass

    def __contains__(self, item):
        item = Expression(item)
        if item.is_constant() and np.all(item.value >= 0):
            return True
        con = PrimalSageCone(item, self.alpha, name='check_mem')
        prob = Problem(CL_MAX, Expression([0]), [con])
        status, value = prob.solve(verbose=False)
        from sageopt.coniclifts import clear_variable_indices
        clear_variable_indices()
        if status == CL_SOLVED:
            return abs(value) < 1e-7
        else:
            return False


class DualSageCone(SetMembership):

    def __init__(self, v, alpha, name, c=None, expcovers=None):
        """
        Aggregrates constraints on "v" so that "v" can be viewed as a dual variable
        to a constraint of the form "c \in C_{SAGE}(alpha)".

        :param v: a Coniclifts Expression of length s.m.
        """
        if c is None:
            self.c = None
        else:
            self.c = Expression(c)
        self.alpha = alpha
        self.ech = ExpCoverHelper(self.alpha, self.c, expcovers)
        self.m = alpha.shape[0]
        self.n = alpha.shape[1]
        self.v = v
        self.name = name
        self.mu_vars = dict()
        self.relent_epi_vars = dict()
        self._variables = self.v.variables()
        self._initialize_variables()
        pass

    def _initialize_variables(self):
        if self.m > 2:
            for i in self.ech.U_I:
                num_cover = np.count_nonzero(self.ech.expcovers[i])
                if num_cover == 0:
                    continue

                var_name = 'mu[' + str(i) + ']_{' + self.name + '}'
                self.mu_vars[i] = Variable(shape=(self.n,), name=var_name)
                self._variables.append(self.mu_vars[i])

                var_name = '_relent_epi_[' + str(i) + ']_{' + self.name + '}'
                epi = Variable(shape=(num_cover,), name=var_name)
                self.relent_epi_vars[i] = epi
                self._variables.append(epi)
        pass

    def variables(self):
        return self._variables

    def conic_form(self):
        nontrivial_I = list(set(self.ech.U_I + self.ech.P_I))
        con = self.v[nontrivial_I] >= 0
        con.epigraph_checked = True
        A_vals, A_rows, A_cols, b, K = con.conic_form()
        cone_data = [(A_vals, A_rows, A_cols, b, K)]
        if self.m > 2:
            for i in self.ech.U_I:
                num_cover = np.count_nonzero(self.ech.expcovers[i])
                if num_cover == 0:
                    continue
                curr_age = self._dual_age_cone_data(i)
                cone_data += curr_age
        return cone_data

    def _dual_age_cone_data(self, i):
        cone_data = []
        selector = self.ech.expcovers[i]
        len_sel = np.count_nonzero(selector)
        #
        # relative entropy constraints
        #
        expr1 = np.tile(self.v[i], len_sel).view(Expression)
        epi = self.relent_epi_vars[i]
        A_vals, A_rows, A_cols, b, K, epi = elementwise_relent(expr1, self.v[selector], epi)
        cone_data.append((A_vals, A_rows, A_cols, b, K))
        #
        # Linear inequalities
        #
        mat = self.alpha[selector, :] - self.alpha[i, :]
        A_vals, A_rows, A_cols = compiled_aff.mat_times_vecvar_minus_vecvar(-mat, self.mu_vars[i], epi)
        num_rows = mat.shape[0]
        b = np.zeros(num_rows)
        K = [Cone('+', num_rows)]
        A_rows = np.array(A_rows)
        cone_data.append((A_vals, A_rows, A_cols, b, K))
        return cone_data

    def _dual_age_cone_violation(self, i, norm_ord, rough, v):
        from sageopt.coniclifts import clear_variable_indices
        selector = self.ech.expcovers[i]
        len_sel = np.count_nonzero(selector)
        if len_sel > 0:
            expr1 = np.tile(v[i], len_sel).ravel()
            expr2 = v[selector].ravel()
            lowerbounds = special_functions.rel_entr(expr1, expr2)
            mat = -(self.alpha[selector, :] - self.alpha[i, :])
            vec = self.mu_vars[i].value
            # compute rough violation for this dual AGE cone
            residual = mat @ vec - lowerbounds
            residual[residual >= 0] = 0
            curr_viol = np.linalg.norm(residual, ord=norm_ord)
            # as applicable, solve an optimization problem to compute the violation.
            if curr_viol > 0 and not rough:
                temp_var = Variable(shape=(mat.shape[1],), name='temp_var')
                prob = Problem(CL_MAX, Expression([0]), [mat @ temp_var >= lowerbounds])
                status, value = prob.solve(verbose=False)
                clear_variable_indices()
                if status == CL_SOLVED and abs(value) < 1e-7:
                    curr_viol = 0
            return curr_viol
        else:
            residual = -v[i] if v[i] < 0 else 0
            return residual

    def violation(self, norm_ord=np.inf, rough=False):
        v = self.v.value
        viols = [0]
        for i in self.ech.U_I:
            num_cover = np.count_nonzero(self.ech.expcovers[i])
            if num_cover == 0:
                continue
            curr_viol = self._dual_age_cone_violation(i, norm_ord, rough, v)
            viols.append(curr_viol)
        viol = max(viols)
        return viol

    def __contains__(self, item):
        from sageopt.coniclifts import clear_variable_indices
        if isinstance(item, Expression):
            item = item.value
        for i in self.ech.U_I:
            selector = self.ech.expcovers[i]
            len_sel = np.count_nonzero(selector)
            expr1 = np.tile(item[i], len_sel)
            expr2 = item[selector]
            lowerbounds = special_functions.rel_entr(expr1, expr2).reshape((-1, 1))
            mat = -(self.alpha[selector, :] - self.alpha[i, :])
            temp_var = Variable(shape=(mat.shape[1], 1), name='temp_var')
            prob = Problem(CL_MAX, Expression([0]), [mat @ temp_var >= lowerbounds])
            status, value = prob.solve(verbose=False)
            clear_variable_indices()
            if status != CL_SOLVED or abs(value) > 1e-7:
                return False
        return True


class ExpCoverHelper(object):

    def __init__(self, alpha, c, expcovers=None):
        if c is not None and not isinstance(c, Expression):
            raise RuntimeError()
        self.alpha = alpha
        self.m = alpha.shape[0]
        self.c = c
        if self.c is not None:
            self.U_I = [i for i, c_i in enumerate(self.c) if (not c_i.is_constant()) or (c_i.offset < 0)]
            # ^ indices of not-necessarily-positive sign; i \in U_I must get an AGE cone.
            # this AGE cone might not be used in the final solution to the associated
            # optimization problem.
            self.N_I = [i for i, c_i in enumerate(self.c) if (c_i.is_constant()) and (c_i.offset < 0)]
            # ^ indices of definitively-negative sign. if j \in N_I, then there is only one AGE
            # cone (indexed by i) with c^{(i)}_j != 0, and that's j == i. These AGE cones will
            # be used in any solution to the associated optimization problem, and we can be
            # certain that c^{(i)}_i == c_i.
            self.P_I = [i for i, c_i in enumerate(self.c) if (c_i.is_constant()) and (c_i.offset > 0)]
            # ^ indices of positive sign.
            # Together, union(self.P_I, self.U_I) comprise all indices "i" where c[i] is not identically zero.
        else:
            self.U_I = [i for i in range(self.m)]
            self.N_I = []
            self.P_I = []
        if isinstance(expcovers, dict):
            self._verify_exp_covers(expcovers)
        elif expcovers is None:
            expcovers = self._default_exp_covers()
        else:
            raise RuntimeError('Argument "expcovers" must be a dict.')
        self.expcovers = expcovers

    def _verify_exp_covers(self, expcovers):
        for i in self.U_I:
            if i not in expcovers:
                raise RuntimeError('Required key missing from "expcovers".')
            if (not isinstance(expcovers[i], np.ndarray)) or (expcovers[i].dtype != bool):
                raise RuntimeError('A value in "expcovers" was the wrong datatype.')
            if expcovers[i][i]:
                warnings.warn('Nonsensical value in "expcovers"; correcting.')
                expcovers[i][i] = False
        pass

    def _default_exp_covers(self):
        expcovers = dict()
        for i in self.U_I:
            cov = np.ones(shape=(self.m,), dtype=bool)
            cov[self.N_I] = False
            cov[i] = False
            expcovers[i] = cov
        # Check if we are likely working with polynomials
        # (if so, go through the trouble of a reduction that
        #  applies to both polynomials and signomials, but
        #  is more likely to be useful in the former case.)
        row_sums = np.sum(self.alpha, 1)
        if np.all(self.alpha >= 0) and np.min(row_sums) == 0:
            # Then apply the reduction.
            zero_loc = np.nonzero(row_sums == 0)[0][0]
            for i in self.U_I:
                if i == zero_loc:
                    continue
                curr_cover = expcovers[i]
                curr_row = self.alpha[i, :]
                for j in range(self.m):
                    if curr_cover[j] and j != zero_loc and curr_row @ self.alpha[j, :] == 0:
                        curr_cover[j] = False
        return expcovers



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
from sageopt import coniclifts as cl
from sageopt.symbolic.signomials import Signomial
from sageopt.relaxations import constraint_generators as con_gen
from sageopt.relaxations import symbolic_correspondences as sym_corr


def primal_sage_cone(sig, name, AbK):
    if AbK is None:
        con = cl.PrimalSageCone(sig.c, sig.alpha, name)
    else:
        A, b, K = AbK
        con = cl.PrimalCondSageCone(sig.c, sig.alpha, A, b, K, name)
    return con


def relative_dual_sage_cone(primal_sig, dual_var, name, AbK):
    if AbK is None:
        con = cl.DualSageCone(dual_var, primal_sig.alpha, name, primal_sig.c)
    else:
        A, b, K = AbK
        con = cl.DualCondSageCone(dual_var, primal_sig.alpha, A, b, K, name, primal_sig.c)
    return con


def sig_relaxation(f, form='dual', ell=0, X=None, mod_supp=None):
    """
    Construct a coniclifts Problem instance for producing a lower bound on

    .. math::

        \min\{ f(x) \,:\, x \\in X \}

    where X = :math:`R^{\\texttt{f.n}}` by default.

    If ``form='dual'``, we can also attempt to recover solutions to the above problem.

    Parameters
    ----------
    f : Signomial
        The objective function to be minimized.
    form : str
        Either ``form='primal'`` or ``form='dual'``.
    ell : int
        The level of the SAGE hierarchy. Must be nonnegative.
    X : dict
        If ``X`` is None, then we produce a bound on ``f`` over :math:`R^{\\texttt{f.n}}`.
        If ``X`` is a dict, then it must contain three fields: ``'AbK'``, ``'gts'``, and ``'eqs'``. For almost all
        applications, the appropriate dict ``X`` can be generated for you by calling ``conditional_sage_data(...)``.
    mod_supp : NumPy ndarray
        This parameter is only used when ``ell > 0``. If ``mod_supp`` is not None, then the rows of this
        array define the exponents of a positive definite modulating Signomial in the SAGE hierarchy.

    Returns
    -------
    prob : sageopt.coniclifts.Problem
        A coniclifts Problem which represents the SAGE relaxation, with given parameters.
        The relaxation can be solved by calling ``prob.solve()``.

    Notes
    -----
    When ``form='primal'``, the Problem can be stated in full generality without too much trouble.
    We define a multiplier signomial ``t`` (with the canonical choice ``t = Signomial(f.alpha, np.ones(f.m))``),
    then return problem data representing ::

        max  gamma
        s.t.    f_mod.c in C_{SAGE}(f_mod.alpha, X)
        where   f_mod := (t ** ell) * (f - gamma).

    Our implementation of Signomial objects allows Variables in the coefficient vector ``c``. As a result, the
    map from ``gamma`` to ``f_mod.c`` is an affine function that takes in a Variable and returns an Expression.
    This makes it very simple to represent ``f_mod.c in C_{SAGE}(f_mod.alpha, X)`` via coniclifts Constraints.
    """
    if form.lower()[0] == 'd':
        prob = sig_dual(f, ell, X, mod_supp)
    elif form.lower()[0] == 'p':
        prob = sig_primal(f, ell, X, None, mod_supp)
    else:
        raise RuntimeError('Unrecognized form: ' + form + '.')
    return prob


def sig_dual(f, ell=0, X=None, modulator_support=None):
    """
    Construct a coniclifts Problem instance for producing a lower bound for ``f`` over the set defined by ``X``,
    and for attempting to recover optimal solutions to the problem ``min{ f(x) | x in X }``.

    Parameters
    ----------
    f : Signomial
        The objective function to be minimized.
    ell : int
        The level of the SAGE hierarchy. Must be nonnegative.
    X : dict
        If ``X`` is None, then we produce a bound on ``f`` over :math:`R^{\\texttt{f.n}}`.
        If ``X`` is a dict, then it must contain three fields: ``'AbK'``, ``'gts'``, and ``'eqs'``. For almost all
        applications, the appropriate dict ``X`` can be generated for you by calling ``conditional_sage_data(...)``.
    modulator_support : NumPy ndarray
        If ``modulator_support`` is not None, then its rows define the exponents of a positive definite
        modulating Signomial in the SAGE hierarchy. This parameter is only used when ``ell > 0``.

    Returns
    -------
    prob : sageopt.coniclifts.Problem
        A coniclifts Problem which represents the dual SAGE relaxation, with given parameters.
        The relaxation can be solved by calling ``prob.solve()``.

    """
    if X is None:
        X = {'AbK': None, 'gts': [], 'eqs': []}
    f.remove_terms_with_zero_as_coefficient()
    # Signomial definitions (for the objective).
    if modulator_support is None:
        modulator_support = f.alpha
    t_mul = Signomial(modulator_support, np.ones(modulator_support.shape[0])) ** ell
    lagrangian = f - cl.Variable(name='gamma')
    metadata = {'f': f, 'lagrangian': lagrangian, 'modulator': t_mul, 'X': X}
    lagrangian = lagrangian * t_mul
    f_mod = f * t_mul
    # C_SAGE^STAR (v must belong to the set defined by these constraints).
    v = cl.Variable(shape=(lagrangian.m, 1), name='v')
    con = relative_dual_sage_cone(lagrangian, v, name='Lagrangian SAGE dual constraint', AbK=X['AbK'])
    constraints = [con]
    # Equality constraint (for the Lagrangian to be bounded).
    a = sym_corr.relative_coeff_vector(t_mul, lagrangian.alpha)
    a = a.reshape(a.size, 1)
    constraints.append(a.T @ v == 1)
    # Objective definition and problem creation.
    obj_vec = sym_corr.relative_coeff_vector(f_mod, lagrangian.alpha)
    obj = obj_vec.T @ v
    # Create coniclifts Problem
    prob = cl.Problem(cl.MIN, obj, constraints)
    prob.associated_data = metadata
    cl.clear_variable_indices()
    return prob


def sig_primal(f, ell=0, X=None, additional_cons=None, modulator_support=None):
    """
    Construct a coniclifts Problem instance for producing a lower bound for ``f`` over the set defined by ``X``.

    Parameters
    ----------
    f : Signomial
        The objective function to be minimized.
    ell : int
        The level of the SAGE hierarchy. Must be nonnegative.
    X : dict
        If ``X`` is None, then we produce a bound on ``f`` over :math:`R^{\\texttt{f.n}}`.
        If ``X`` is a dict, then it must contain three fields: ``'AbK'``, ``'gts'``, and ``'eqs'``. For almost all
        applications, the appropriate dict ``X`` can be generated for you by calling ``conditional_sage_data(...)``.
    modulator_support : NumPy ndarray
        If ``modulator_support`` is not None, then its rows define the exponents of a positive definite
        modulating Signomial in the SAGE hierarchy. This parameter is only used when ``ell > 0``.
    additional_cons: list of coniclifts Constraint objects
        Some primal SAGE polynomial relaxations can easily be transformed to primal SAGE signomial relaxations,
        by way of "signomial representatives". The signomial representatives are often accompanied by additional
        constraints. Those constraints may be passed as a list, via this argument. End-users are unlikely to use
        this argument.

    Returns
    -------
    prob : sageopt.coniclifts.Problem
        A coniclifts Problem which represents the primal SAGE relaxation, with given parameters.
        The relaxation can be solved by calling ``prob.solve()``.

    Notes
    -----
    Unlike ``sig_dual``, the ``sig_primal`` formulation can be stated in full generality without too much trouble.
    We define a multiplier signomial ``t`` (with the canonical choice ``t = Signomial(f.alpha, np.ones(f.m))``),
    then return problem data representing ::

        max  gamma
        s.t.    f_mod.c in C_{SAGE}(f_mod.alpha, X)
        where   f_mod := (t ** ell) * (f - gamma).

    Our implementation of Signomial objects allows Variables in the coefficient vector ``c``. As a result, the
    map from ``gamma`` to ``f_mod.c`` is an affine function that takes in a Variable and returns an Expression.
    This makes it very simple to represent ``f_mod.c in C_{SAGE}(f_mod.alpha, X)`` via coniclifts Constraints.
    """
    if X is None:
        X = {'AbK': None, 'gts': [], 'eqs': []}
    f.remove_terms_with_zero_as_coefficient()
    if modulator_support is None:
        modulator_support = f.alpha
    t = Signomial(modulator_support, np.ones(modulator_support.shape[0]))
    gamma = cl.Variable(name='gamma')
    s_mod = (f - gamma) * (t ** ell)
    s_mod.remove_terms_with_zero_as_coefficient()
    con = primal_sage_cone(s_mod, name=str(s_mod), AbK=X['AbK'])
    constraints = [con]
    obj = gamma.as_expr()
    if additional_cons is not None:
        constraints += additional_cons
    prob = cl.Problem(cl.MAX, obj, constraints)
    cl.clear_variable_indices()
    return prob


def sage_feasibility(f, X=None, additional_cons=None):
    """
    Constructs a coniclifts maximization Problem which is feasible iff ``f`` admits an X-SAGE decomposition.

    Parameters
    ----------
    f : Signomial
        We want to test if this function admits an X-SAGE decomposition.
    X : dict
        If ``X`` is None, then we test nonnegativity of ``f`` over :math:`R^{\\texttt{f.n}}`.
        If ``X`` is a dict, then it must contain three fields: ``'AbK'``, ``'gts'``, and ``'eqs'``. For almost all
        applications, the appropriate dict ``X`` can be generated for you by calling ``conditional_sage_data(...)``.
    additional_cons : :obj:`list` of :obj:`sageopt.coniclifts.Constraint`
        This is mostly used for SAGE polynomials. When provided, it should be a list of Constraints over
        coniclifts Variables appearing in ``f.c``.

    Returns
    -------
    prob : sageopt.coniclifts.Problem
        A coniclifts maximization Problem. If ``f`` admits an X-SAGE decomposition, then we should have
        ``prob.value > -np.inf``, once ``prob.solve()`` has been called.
    """
    if X is None:
        X = {'AbK': None, 'gts': [], 'eqs': []}
    f.remove_terms_with_zero_as_coefficient()
    con = primal_sage_cone(f, name=str(f), AbK=X['AbK'])
    constraints = [con]
    if additional_cons is not None:
        constraints += additional_cons
    prob = cl.Problem(cl.MAX, cl.Expression([0]), constraints)
    cl.clear_variable_indices()
    return prob


def sage_multiplier_search(f, level=1, X=None):
    """
    Constructs a coniclifts maximization Problem which is feasible if ``f`` can be certified as nonnegative
    over ``X``, by using an appropriate X-SAGE modulating function.

    Parameters
    ----------
    f : Signomial
        We want to test if ``f`` is nonnegative over ``X``.
    level : int
        Controls the complexity of the X-SAGE modulating function. Must be a positive integer.
    X : dict
        If ``X`` is None, then we test nonnegativity of ``f`` over :math:`R^{\\texttt{f.n}}`.
        If ``X`` is a dict, then it must contain three fields: ``'AbK'``, ``'gts'``, and ``'eqs'``. For almost all
        applications, the appropriate dict ``X`` can be generated for you by calling ``conditional_sage_data(...)``.

    Returns
    -------
    prob : sageopt.coniclifts.Problem

    Notes
    -----
    This function provides an alternative to moving up the SAGE hierarchy, for the goal of certifying
    nonnegativity of a signomial ``f`` over some convex set ``X``.  In general, the approach is to introduce
    a signomial

        ``mult = Signomial(alpha_hat, c_tilde)``

    where the rows of ``alpha_hat`` are all ``level``-wise sums of rows from ``f.alpha``, and ``c_tilde``
    is a coniclifts Variable defining a nonzero X-SAGE function. Then we check if ``f_mod = f * mult``
    is X-SAGE for any choice of ``c_tilde``.
    """
    if X is None:
        X = {'AbK': None, 'gts': [], 'eqs': []}
    f.remove_terms_with_zero_as_coefficient()
    constraints = []
    mult_alpha = hierarchy_e_k([f], k=level)
    c_tilde = cl.Variable(mult_alpha.shape[0], name='c_tilde')
    mult = Signomial(mult_alpha, c_tilde)
    constraints.append(cl.sum(c_tilde) >= 1)
    sig_under_test = mult * f
    con1 = primal_sage_cone(mult, name=str(mult), AbK=X['AbK'])
    con2 = primal_sage_cone(sig_under_test, name=str(sig_under_test), AbK=X['AbK'])
    constraints.append(con1)
    constraints.append(con2)
    prob = cl.Problem(cl.MAX, cl.Expression([0]), constraints)
    cl.clear_variable_indices()
    return prob


def sig_constrained_relaxation(f, gts, eqs, form='dual', p=0, q=1, ell=0, X=None):
    """
    Construct a coniclifts Problem instance representing a level-``(p, q, ell)`` SAGE relaxation
    for the signomial program

    .. math::

        \\begin{align*}
          \min\{ f(x) :~& g(x) \geq 0 \\text{ for } g \\in \\text{gts}, \\\\
                       & g(x) = 0  \\text{ for } g \\in \\text{eqs}, \\\\
                       & \\text{and } x \\in X \}
        \\end{align*}

    where X = :math:`R^{\\texttt{f.n}}` by default. The optimal value of this relaxation will produce
    a lower bound on the minimization problem described above. When ``form='dual'``,
    a solution to this relaxation can be used to help recover optimal solutions to
    the problem described above.

    Parameters
    ----------
    f : Signomial
        The objective function to be minimized.
    gts : list of Signomials
        For every ``g in gts``, there is a desired constraint that variables ``x`` satisfy ``g(x) >= 0``.
    eqs : list of Signomials
        For every ``g in eqs``, there is a desired constraint that variables ``x`` satisfy ``g(x) == 0``.
    form : str
        Either ``form='primal'`` or ``form='dual'``.
    p : int
        Controls the complexity of Lagrange multipliers in the primal formulation, and (equivalently) constraints in
        the dual formulation. The smallest value is ``p=0``, which corresponds to scalar Lagrange multipliers.
    q : int
        The number of folds applied to the constraints ``gts`` and ``eqs``. The smallest value is ``q=1``, which
        means "leave ``gts`` and ``eqs`` as-is."
    ell : int
        Controls the complexity of any modulator applied to the Lagrangian in the primal formulation, and
        (equivalently) constraints in the dual formulation. The smallest value is ``ell=0``, which means
        the primal Lagrangian must be a SAGE signomial.
    X : dict
        If ``X`` is None, then this parameter is ignored.
        If ``X`` is a dict, then it must contain three fields: ``'AbK'``, ``'gts'``, and ``'eqs'``. For almost all
        applications, the appropriate dict ``X`` can be generated for you by calling ``conditional_sage_data(...)``.

    Returns
    -------
    prob : sageopt.coniclifts.Problem

    Notes
    -----
    The exact meaning of parameters ``p`` and ``q`` is determined by the function ``make_lagrangian(...).``.

    The meaning of the parameter ``ell`` is most clear from the implementation of ``sig_constrained_primal(...)``,
    although both ``sig_constrained_primal`` and ``sig_constrained_dual`` contain logic for handling this parameter.

    """
    if form.lower()[0] == 'd':
        prob = sig_constrained_dual(f, gts, eqs, p, q, ell, X)
    elif form.lower()[0] == 'p':
        prob = sig_constrained_primal(f, gts, eqs, p, q, ell, X)
    else:
        raise RuntimeError('Unrecognized form: ' + form + '.')
    return prob
    pass


def sig_constrained_primal(f, gts, eqs, p=0, q=1, ell=0, X=None):
    """
    Construct the SAGE-(p, q, ell) primal problem for the signomial program

        min{ f(x) : g(x) >= 0 for g in gts,
                    g(x) == 0 for g in eqs,
                    and x in X }

    where X = :math:`R^{\\texttt{f.n}}` by default.

    Parameters
    ----------
    f : Signomial
        The objective function to be minimized.
    gts : list of Signomials
        For every ``g in gts``, there is a desired constraint that variables ``x`` satisfy ``g(x) >= 0``.
    eqs : list of Signomials
        For every ``g in eqs``, there is a desired constraint that variables ``x`` satisfy ``g(x) == 0``.
    p : int
        Controls the complexity of Lagrange multipliers. Smallest value is ``p=0``, which corresponds to scalars.
    q : int
        The number of folds applied to the constraints ``gts`` and ``eqs``. Smallest value is ``q=1``, which means
        "leave ``gts`` and ``eqs`` as-is."
    ell : int
        Controls the complexity of any modulator applied to the Lagrangian.
        The smallest value is ``ell=0``, which means that the Lagrangian must be SAGE.
        ``ell=1`` means ``tilde_L = modulator * Lagrangian`` must be SAGE.
    X : dict
        If ``X`` is None, then this parameter is ignored.
        If ``X`` is a dict, then it must contain three fields: ``'AbK'``, ``'gts'``, and ``'eqs'``. For almost all
        applications, the appropriate dict ``X`` can be generated for you by calling ``conditional_sage_data(...)``.

    Returns
    -------
    prob : sageopt.coniclifts.Problem

    """
    if X is None:
        X = {'AbK': None, 'gts': [], 'eqs': []}
    lagrangian, ineq_lag_mults, _, gamma = make_sig_lagrangian(f, gts, eqs, p=p, q=q)
    metadata = {'lagrangian': lagrangian, 'X': X}
    if ell > 0:
        alpha_E_1 = hierarchy_e_k([f] + list(gts) + list(eqs), k=1)
        modulator = Signomial(alpha_E_1, np.ones(alpha_E_1.shape[0])) ** ell
        lagrangian = lagrangian * modulator
    else:
        modulator = Signomial({(0,) * f.n: 1})
    metadata['modulator'] = modulator
    # The Lagrangian (after possible multiplication, as above) must be a SAGE signomial.
    con = primal_sage_cone(lagrangian, name='Lagrangian is SAGE', AbK=X['AbK'])
    constrs = [con]
    #  Lagrange multipliers (for inequality constraints) must be SAGE signomials.
    for i, (s_h, _) in enumerate(ineq_lag_mults):
        con_name = 'SAGE multiplier for signomial inequality # ' + str(i)
        con = primal_sage_cone(s_h, name=con_name, AbK=X['AbK'])
        constrs.append(con)
    # Construct the coniclifts Problem.
    prob = cl.Problem(cl.MAX, gamma, constrs)
    prob.associated_data = metadata
    cl.clear_variable_indices()
    return prob


def sig_constrained_dual(f, gts, eqs, p=0, q=1, ell=0, X=None):
    """
    Construct the SAGE-(p, q, ell) dual problem for the signomial program

        min{ f(x) : g(x) >= 0 for g in gts,
                    g(x) == 0 for g in eqs,
                    and x in X }

    where X = :math:`R^{\\texttt{f.n}}` by default.

    Parameters
    ----------
    f : Signomial
        The objective function to be minimized.
    gts : list of Signomials
        For every ``g in gts``, there is a desired constraint that variables ``x`` satisfy ``g(x) >= 0``.
    eqs : list of Signomials
        For every ``g in eqs``, there is a desired constraint that variables ``x`` satisfy ``g(x) == 0``.
    p : int
        Controls the complexity of Lagrange multipliers in the primal problem, and in an turn the complexity of
        constraints in this dual problem. The smallest value is ``p=0``, which corresponds to scalar Lagrange
        multipliers in the primal problem, and linear inequality constraints in this dual problem.
    q : int
        The number of folds applied to the constraints ``gts`` and ``eqs``. Smallest value is ``q=1``, which means
        "leave ``gts`` and ``eqs`` as-is."
    ell : int
        Controls the complexity of any modulator applied to the Lagrangian in the primal problem.
        The smallest value is ``ell=0``, which means the primal Lagrangian is unchanged / not modulated.
    X : dict
        If ``X`` is None, then this parameter is ignored.
        If ``X`` is a dict, then it must contain three fields: ``'AbK'``, ``'gts'``, and ``'eqs'``. For almost all
        applications, the appropriate dict ``X`` can be generated for you by calling ``conditional_sage_data(...)``.

    Returns
    -------
    prob : sageopt.coniclifts.Problem

    """
    if X is None:
        X = {'AbK': None, 'gts': [], 'eqs': []}
    lagrangian, ineq_lag_mults, eq_lag_mults, _ = make_sig_lagrangian(f, gts, eqs, p=p, q=q)
    metadata = {'lagrangian': lagrangian, 'f': f, 'gts': gts, 'eqs': eqs, 'level': (p, q, ell), 'X': X}
    if ell > 0:
        alpha_E_1 = hierarchy_e_k([f] + list(gts) + list(eqs), k=1)
        modulator = Signomial(alpha_E_1, np.ones(alpha_E_1.shape[0])) ** ell
        lagrangian = lagrangian * modulator
        f = f * modulator
    else:
        modulator = Signomial({(0,) * f.n: 1})
    metadata['modulator'] = modulator
    # In primal form, the Lagrangian is constrained to be a SAGE signomial.
    # Introduce a dual variable "v" for this constraint.
    v = cl.Variable(shape=(lagrangian.m, 1), name='v')
    con = relative_dual_sage_cone(lagrangian, v, name='Lagrangian SAGE dual constraint', AbK=X['AbK'])
    constraints = [con]
    for i, (s_h, h) in enumerate(ineq_lag_mults):
        # These generalized Lagrange multipliers "s_h" are SAGE signomials.
        # For each such multiplier, introduce an appropriate dual variable "v_h", along
        # with constraints over that dual variable.
        v_h = cl.Variable(name='v_' + str(h), shape=(s_h.m, 1))
        con_name = 'SAGE dual for signomial inequality # ' + str(i)
        con = relative_dual_sage_cone(s_h, v_h, name=con_name, AbK=X['AbK'])
        constraints.append(con)
        h = h * modulator
        c_h = sym_corr.moment_reduction_array(s_h, h, lagrangian)
        constraints.append(c_h @ v == v_h)
    for s_h, h in eq_lag_mults:
        # These generalized Lagrange multipliers "s_h" are arbitrary signomials.
        # They dualize to homogeneous equality constraints.
        h = h * modulator
        c_h = sym_corr.moment_reduction_array(s_h, h, lagrangian)
        constraints.append(c_h @ v == 0)
    # Equality constraint (for the Lagrangian to be bounded).
    a = sym_corr.relative_coeff_vector(modulator, lagrangian.alpha)
    constraints.append(a.T @ v == 1)
    # Define the dual objective function.
    obj_vec = sym_corr.relative_coeff_vector(f, lagrangian.alpha)
    obj = obj_vec.T @ v
    # Return the coniclifts Problem.
    prob = cl.Problem(cl.MIN, obj, constraints)
    prob.associated_data = metadata
    cl.clear_variable_indices()
    return prob


def make_sig_lagrangian(f, gts, eqs, p, q):
    """
    Given a problem

    .. math::

        \\begin{align*}
          \min\{ f(x) :~& g(x) \geq 0 \\text{ for } g \\in \\text{gts}, \\\\
                       & g(x) = 0  \\text{ for } g \\in \\text{eqs}, \\\\
                       & \\text{and } x \\in X \}
        \\end{align*}

    construct the q-fold constraints ``q-gts`` and ``q-eqs,`` and the Lagrangian

    .. math::

        L = f - \\gamma
            - \sum_{g \, \\in  \, \\text{q-gts}} s_g \cdot g
            - \sum_{g \, \\in  \, \\text{q-eqs}} z_g \cdot g

    where :math:`\\gamma` and the coefficients on Signomials :math:`s_g` and :math:`z_g`
    are coniclifts Variables.

    Parameters
    ----------
    f : Signomial
        The objective in a desired minimization problem.
    gts : list of Signomials
        For every ``g in gts``, there is a desired constraint that variables ``x`` satisfy ``g(x) >= 0``.
    eqs : list of Signomials
        For every ``g in eqs``, there is a desired constraint that variables ``x`` satisfy ``g(x) == 0``.
    p : int
        Controls the complexity of ``s_g`` and ``z_g``.
    q : int
        The number of folds of constraints ``gts`` and ``eqs``.

    Returns
    -------
    L : Signomial
        ``L.c`` is an affine expression of coniclifts Variables.

    ineq_dual_sigs : a list of pairs of Signomial objects.
        If the pair ``(s1, s2)`` is in this list, then ``s1`` is a generalized Lagrange multiplier
        to the constraint ``s2(x) >= 0``.

    eq_dual_sigs : a list of pairs of Signomial objects.
        If the pair ``(s1, s2)`` is in this list, then ``s1`` is a generalized Lagrange multiplier to the
        constraint ``s2(x) == 0``. This return value is not accessed for primal-form SAGE relaxations.

    gamma : coniclifts.Variable.
        In primal-form SAGE relaxations, we want to maximize ``gamma``. In dual form SAGE relaxations,
        ``gamma`` induces a normalizing equality constraint. This return value is not accessed for
        dual-form SAGE relaxations.


    Notes
    -----
    The values returned by this function are used to construct constrained SAGE relaxations.
    The basic primal SAGE relaxation is obtained by maximizing gamma, subject to the constraint
    that ``L`` and each ``s_g`` are SAGE functions. The dual SAGE relaxation is obtained by
    symbolically applying conic duality to the primal.
    """
    folded_gt = con_gen.up_to_q_fold_cons(gts, q)
    gamma = cl.Variable(name='gamma')
    L = f - gamma
    alpha_E_p = hierarchy_e_k([f] + list(gts) + list(eqs), k=p)
    ineq_dual_sigs = []
    for g in folded_gt:
        s_g_coeff = cl.Variable(name='s_' + str(g), shape=(alpha_E_p.shape[0],))
        s_g = Signomial(alpha_E_p, s_g_coeff)
        L -= s_g * g
        ineq_dual_sigs.append((s_g, g))
    eq_dual_sigs = []
    folded_eq = con_gen.up_to_q_fold_cons(eqs, q)
    for g in folded_eq:
        z_g_coeff = cl.Variable(name='z_' + str(g), shape=(alpha_E_p.shape[0],))
        z_g = Signomial(alpha_E_p, z_g_coeff)
        L -= z_g * g
        eq_dual_sigs.append((z_g, g))
    return L, ineq_dual_sigs, eq_dual_sigs, gamma


def conditional_sage_data(f, gts, eqs):
    """
    Identify a subset of the constraints in ``gts`` and ``eqs`` which can be incorporated into
    conditional SAGE relaxations. Generate conic data that relaxation-constructors will need
    in downstream applications.

    Parameters
    ----------
    f : Signomial
        The objective in a desired optimization problem. This parameter is only used to determine
        the dimension of the set defined by constraints in ``gts`` and ``eqs``.
    gts : list of Signomials
        For every ``g in gts``, there is a desired constraint that variables ``x`` satisfy ``g(x) >= 0``.
    eqs : list of Signomial
        For every ``g in eqs``, there is a desired constraint that variables ``x`` satisfy ``g(x) == 0``.

    Returns
    -------
    X : dict

        ``X`` is keyed by three strings: ``'AbK'``, ``'gts'``, and ``'eqs'``.

        ``X['gts']`` is a list of Signomials so that every ``g in X['gts']`` has an efficient
        convex representation of ``{x : g(x) >= 0}``. The intersection of all of these sets
        is contained within ``{x : g(x) >= 0 for all g in gts}``. ``X['eqs']`` is defined
        similarly, but for equality constraints instead of inequality constraints.
        ``X['AbK']`` is a conic representation of the feasible set cut out by
        ``X['gts']`` and ``X['eqs']``. The conic representation is a triple ``X['AbK'] = (A, b, K)``,
        where ``A`` is a SciPy sparse matrix, ``b`` is a numpy 1d array, and ``K`` is a list of
        coniclifts Cone objects. The number of columns for ``A`` in ``X['AbK']`` will always
        be at least ``f.n``. If the number of columns is greater than ``f.n``, then the first ``f.n``
        columns of ``A`` correspond to the variables over which ``f`` is defined. Any
        remaining columns are auxiliary variables needed to represent ``X`` in coniclifts primitives.

    Notes
    -----
        This function essentially defines the requirements for ``X`` which may be passed to
        conditional SAGE relaxations defined in this python module.

        If a user wants to define their own ``X`` without calling this function, that is doable.
        A possible use-case here is when ``X`` has a simple convex description, but not a simple
        description in terms of signomial inequality and equality constraints.

        Suppose for example that we want ``X`` to represent the ell-2 unit ball in R^{f.n}.
        This can be accomplished as follows ::

            import sageopt.coniclifts as cl
            import numpy as np
            x = cl.Variable(shape=(f.n,), name='x')
            constraints = [1 >= cl.vector2norm(x)]
            A, b, K, _, _ = cl.compile_constrained_system(constraints)
            cl.clear_variable_indices()
            my_gts = [lambda dummy_x: 1 - np.linalg.norm(dummy_x, ord=2)]
            my_eqs = []
            X = {'AbK': (A, b, K), 'gts': my_gts, 'eqs': my_eqs}

        The message of this example is that ``X['gts']`` and ``X['eqs']`` don't actually need
        to be lists of Signomials. They just need to be callable functions which define membership
        in ``X['AbK]``. To really hit this home, functions in ``X['gts']`` don't even need to be
        continuous! For example, if you wanted to test membership in the ell-2 ball without
        allowing even the smallest constraint violation, you could use the function ::

            my_gts = [lambda dummy_x: 1 if 1 >= np.linalg.norm(dummy_x, ord=2) else -np.inf]

        and every important feature of sageopt's signomial relaxations would work as expected.
    """
    x = cl.Variable(shape=(f.n,), name='x')
    coniclift_cons = []
    conv_gt = con_gen.valid_posynomial_inequalities(gts)
    for g in conv_gt:
        nonconst_selector = np.ones(shape=(g.m,), dtype=bool)
        nonconst_selector[g.constant_location()] = False
        if g.m > 2:
            cst = g.c[~nonconst_selector]
            alpha = g.alpha[nonconst_selector, :]
            c = -g.c[nonconst_selector]
            expr = cl.weighted_sum_exp(c, alpha @ x)
            coniclift_cons.append(expr <= cst)
        elif g.m == 2:
            expr = g.alpha[nonconst_selector, :] @ x
            cst = np.log(g.c[~nonconst_selector] / abs(g.c[nonconst_selector]))
            coniclift_cons.append(expr <= cst)
        else:
            raise RuntimeError('Trivial (or infeasible) signomial constraint.')
    conv_eqs = con_gen.valid_monomial_equations(eqs)
    for g in conv_eqs:
        # g is of the form c1 - c2 * exp(a.T @ x) == 0, where c1, c2 > 0
        cst_loc = g.constant_location()
        non_cst_loc = 1 if cst_loc == 0 else 1
        rhs = np.log(g.c[cst_loc] / abs(g.c[non_cst_loc]))
        coniclift_cons.append(g.alpha[non_cst_loc, :] @ x == rhs)
    if len(coniclift_cons) > 0:
        A, b, K, _, _ = cl.compile_constrained_system(coniclift_cons)
        cl.clear_variable_indices()
        AbK = (A, b, K)
        X = {'AbK': AbK, 'gts': conv_gt, 'eqs': conv_eqs}
        return X
    else:
        cl.clear_variable_indices()
        X = {'AbK': None, 'gts': [], 'eqs': []}
        return X


def hierarchy_e_k(sigs, k):
    alpha_tups = sum([list(s.alpha_c.keys()) for s in sigs], [])
    alpha_tups = set(alpha_tups)
    s = Signomial(dict([(a, 1.0) for a in alpha_tups]))
    s = s ** k
    return s.alpha



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""


class Cone(object):

    def __init__(self, cone_type, length, annotations=None):
        self.type = cone_type
        self.len = length
        if annotations is None:
            annotations = dict()
        self.annotations = annotations

    def __eq__(self, other):
        if isinstance(other, Cone):
            return self.type == other.type and self.len == other.len and self.annotations == other.annotations
        else:
            return False

    @staticmethod
    def annotate_cone_positions(K):
        running_idx = 0
        for i, co in enumerate(K):
            co.annotations['ss'] = (running_idx, running_idx + co.len)
            co.annotations['position'] = i
            running_idx += co.len
        pass

    @staticmethod
    def annotate_cone_scopes(A, K):
        # assumes entries of K have already been annotated with position data.
        for co in K:
            co_start, co_stop = co.annotations['ss']
            if 'A' in co.annotations:
                curr_A = co.annotations['A']
            else:
                curr_A = A[co_start:co_stop, :]
            co.annotations['scope'] = {i for i in curr_A.nonzero()[1]}
        pass








"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
from sageopt.coniclifts.constraints.set_membership.setmem import SetMembership
from sageopt.coniclifts.cones import Cone
from sageopt.coniclifts.base import Expression, ScalarVariable, Variable
from sageopt.coniclifts.operators.norms import vector2norm
from sageopt.coniclifts.problems.problem import Problem


class PrimalProductCone(SetMembership):

    _CONSTRAINT_ID_ = 0

    def __init__(self, y, K):
        self.id = PrimalProductCone._CONSTRAINT_ID_
        PrimalProductCone._CONSTRAINT_ID_ += 1
        if not isinstance(y, Expression):
            y = Expression(y)
        K_size = sum([co.len for co in K])
        if not y.size == K_size:
            raise RuntimeError('Incompatible dimensions for y (' + str(y.size) + ') and K (' + str(K_size) + ')')
        if np.any([co.type == 'P' for co in K]):
            raise RuntimeError('This function does not currently support the PSD cone.')
        self.y = y.ravel()
        self.K = K
        self.K_size = K_size
        pass

    def variables(self):
        return self.y.variables()

    def conic_form(self):
        A_rows, A_cols, A_vals = [], [], []
        b = np.zeros(shape=(self.K_size,))
        for i, se in enumerate(self.y.flat):
            if len(se.atoms_to_coeffs) == 0:
                b[i] = se.offset
                A_rows.append(i)
                A_cols.append(int(ScalarVariable.curr_variable_count()) - 1)
                A_vals.append(0)  # make sure scipy infers correct dimensions later on.
            else:
                b[i] = se.offset
                A_rows += [i] * len(se.atoms_to_coeffs)
                col_idx_to_coeff = [(a.id, c) for a, c in se.atoms_to_coeffs.items()]
                A_cols += [atom_id for (atom_id, _) in col_idx_to_coeff]
                A_vals += [c for (_, c) in col_idx_to_coeff]
        return [(A_vals, np.array(A_rows), A_cols, b, self.K)]

    @staticmethod
    def project(item, K):
        from sageopt.coniclifts import MIN as CL_MIN, clear_variable_indices
        item = Expression(item).ravel()
        x = Variable(shape=(item.size,))
        t = Variable(shape=(1,))
        cons = [vector2norm(item - x) <= t, PrimalProductCone(x, K)]
        prob = Problem(CL_MIN, t, cons)
        prob.solve(verbose=False)
        clear_variable_indices()
        return prob.value

    def __contains__(self, item, tol=1e-7):
        residual = PrimalProductCone.project(item, self.K)
        return residual < tol


class DualProductCone(SetMembership):

    _CONSTRAINT_ID_ = 0

    def __init__(self, y, K):
        # y must belong to K^dagger
        self.id = DualProductCone._CONSTRAINT_ID_
        DualProductCone._CONSTRAINT_ID_ += 1
        if not isinstance(y, Expression):
            y = Expression(y)
        K_size = sum([co.len for co in K])
        if not y.size == K_size:
            raise RuntimeError('Incompatible dimensions for y (' + str(y.size) + ') and K (' + str(K_size) + ')')
        if np.any([co.type == 'P' for co in K]):
            raise RuntimeError('This function does not currently support the PSD cone.')
        self.y = y.ravel()
        self.K = K
        self.K_size = K_size
        pass

    def variables(self):
        return self.y.variables()

    def conic_form(self):
        self_dual_cones = {'+', 'S', 'P'}
        start_row = 0
        y_mod = []
        for co in self.K:
            stop_row = start_row + co.len
            if co.type in self_dual_cones:
                y_mod.append(self.y[start_row:stop_row])
            elif co.type == 'e':
                temp_y = np.array([-self.y[start_row + 2],
                                   np.exp(1) * self.y[start_row + 1],
                                   -self.y[start_row]])
                y_mod.append(temp_y)
            elif co.type != '0':
                raise RuntimeError('Unexpected cone type (' + str(co.type) + ').')
            start_row = stop_row
        y_mod = np.hstack(y_mod)
        y_mod = Expression(y_mod)
        # Now we can pretend all nonzero cones are self-dual.
        A_vals, A_rows, A_cols = [], [], []
        cur_K = [Cone(co.type, co.len) for co in self.K if co.type != '0']
        cur_K_size = sum([co.len for co in cur_K])
        if cur_K_size > 0:
            b = np.zeros(shape=(cur_K_size,))
            for i, se in enumerate(y_mod):
                if len(se.atoms_to_coeffs) == 0:
                    b[i] = se.offset
                    A_rows.append(i)
                    A_cols.append(int(ScalarVariable.curr_variable_count()) - 1)
                    A_vals.append(0)  # make sure scipy infers correct dimensions later on.
                else:
                    b[i] = se.offset
                    A_rows += [i] * len(se.atoms_to_coeffs)
                    col_idx_to_coeff = [(a.id, c) for a, c in se.atoms_to_coeffs.items()]
                    A_cols += [atom_id for (atom_id, _) in col_idx_to_coeff]
                    A_vals += [c for (_, c) in col_idx_to_coeff]
            return [(A_vals, np.array(A_rows), A_cols, b, cur_K)]
        else:
            return [([], np.zeros(shape=(0,), dtype=int), [], np.zeros(shape=(0,)), [])]

    @staticmethod
    def project(item, K):
        from sageopt.coniclifts import MIN as CL_MIN, clear_variable_indices
        item = Expression(item).ravel()
        x = Variable(shape=(item.size,))
        t = Variable(shape=(1,))
        cons = [vector2norm(item - x) <= t, DualProductCone(x, K)]
        prob = Problem(CL_MIN, t, cons)
        prob.solve(verbose=False)
        clear_variable_indices()
        return prob.value

    def __contains__(self, item, tol=1e-7):
        residual = DualProductCone.project(item, self.K)
        return residual < tol



from sageopt.relaxations import sage_polys
from sageopt.relaxations import sage_sigs
from sageopt.relaxations import constraint_generators

from sageopt.symbolic.signomials import Signomial
from sageopt.symbolic.polynomials import Polynomial

# import function names with the sig_ prefix.
from sageopt.relaxations.sage_sigs import sig_primal, sig_dual, sig_constrained_primal, sig_constrained_dual
from sageopt.relaxations.sage_sigs import sig_relaxation, sig_constrained_relaxation
from sageopt.relaxations.sig_solution_recovery import sig_solrec

# import function names with the poly_ prefix.
from sageopt.relaxations.sage_polys import poly_primal, poly_dual, poly_constrained_primal, poly_constrained_dual
from sageopt.relaxations.sage_polys import poly_relaxation, poly_constrained_relaxation
from sageopt.relaxations.poly_solution_recovery import poly_solrec

# import remaining functions for solution recovery
from sageopt.relaxations.sig_solution_recovery import local_refine, is_feasible
from sageopt.relaxations.poly_solution_recovery import local_refine_polys_from_sigs

# import functions that don't have prefixes, and which this file needs to wrap.
from sageopt.relaxations.sage_sigs import sage_feasibility as _sig_sage_feasibility
from sageopt.relaxations.sage_polys import sage_feasibility as _poly_sage_feasibility
from sageopt.relaxations.sage_sigs import sage_multiplier_search as _sig_sage_mult_search
from sageopt.relaxations.sage_polys import sage_multiplier_search as _poly_sage_mult_search
from sageopt.relaxations.sage_sigs import conditional_sage_data as _sig_cond_sage_data
from sageopt.relaxations.sage_polys import conditional_sage_data as _poly_cond_sage_data


def sage_feasibility(f, X=None):
    """
    Return a coniclifts Problem ``prob`` with ``prob.value > -np.inf`` iff
    ``f`` admits an  X-SAGE decomposition.

    Parameters
    ----------
    f : Signomial or Polynomial
        We want to know if this function admits an X-SAGE decomposition
    X : dict or None
        If ``X`` is a dict, then it must be generated in accordance with the
        function ``conditional_sage_data``.

    Returns
    -------
    A coniclifts Problem object, with objective "maximize 0", and constraints
    requiring that ``f`` is X-SAGE. If the argument ``X`` is ``None``, then
    we test if ``f`` is R^n-SAGE.

    Notes
    -----
    This function is simply a wrapper around two functions of the same name,
    which were written for cases where ``f`` is a Signomial or Polynomial.

    The signomial and polynomial cases are implemented in
    ``sageopt.relaxations.sage_sigs.sage_feasibility`` and
    ``sageopt.relaxations.sage_polys.sage_feasibility`` respectively.

    """
    if isinstance(f, Signomial):
        prob = _sig_sage_feasibility(f, X)
        return prob
    elif isinstance(f, Polynomial):
        prob = _poly_sage_feasibility(f, X)
        return prob
    else:
        raise ValueError('"f" must be a Signomial or Polynomial.')


def sage_multiplier_search(f, level=1, X=None):
    """
    Return a coniclifts Problem ``prob``, where ``prob.value > -np.inf`` iff ``f``
    can be certified as nonnegative over ``X`` (using a modulator of complexity
    determined by ``level``).

    Parameters
    ----------
    f : Signomial or Polynomial
        We are looking for a certificate that ``f`` is nonnegative over ``X``.
    level : int
        A positive integer. This determines the complexity of a modulator ``g``
        which is positive over ``X``.
    X : dict or None
        If ``X`` is a dict, then it must be generated in accordance with the
        function ``conditional_sage_data``. If ``X`` is None, then we certify
        nonnegativity over R^n.

    Returns
    -------
    prob : coniclifts.Problem
        A problem with objective "maximize 0", and constraints that both ``g``
        and ``f * g`` are X-SAGE. The main variables in this problem are the coefficients
        of ``g``. The complexity of ``g`` is determined by ``level``.

    Notes
    -----
    This function is simply a wrapper around two functions of the same name,
    which were written for cases where ``f`` is a Signomial or Polynomial.

    The signomial and polynomial cases are implemented in
    ``sageopt.relaxations.sage_sigs.sage_multiplier_search`` and
    ``sageopt.relaxations.sage_polys.sage_multiplier_search`` respectively.

    """
    if isinstance(f, Polynomial):
        prob = _poly_sage_mult_search(f, level, X)
        return prob
    elif isinstance(f, Signomial):
        prob = _sig_sage_mult_search(f, level, X)
        return prob
    else:
        raise ValueError('"f" must be a Signomial or Polynomial.')


def conditional_sage_data(f, gts, eqs):
    """
    Infer (and construct a representation for) a tractable set ``X`` which
    is contained within { x : g(x) >= 0 for g in gts } and
    {x : g(x) == 0 for g in eqs}. For use in conditional SAGE relaxations.

    Parameters
    ----------
    f : Signomial or Polynomial
        This is only used to determine the dimension of the output ``X``
    gts : list of Signomials, or list of Polynomials
        Desired inequality constraint functions.
    eqs : list of Signomials, or list of Polynomials
        Desired equality constraint functions.

    Returns
    -------
    X : dict
        ``X`` has three key-value pairs.

        ``X['gts']`` is a list of functions which define inequality constraints,
        and ``X['eqs']`` is a list of functions which define equality constraints.

        The third key depends on if provided functions were Signomials or Polynomials.
        If the given functions were Signomials, then ``X`` has a key  ``X['AbK']``.
        If the given functions were Polynomials, then ``X`` has a key ``X['log_AbK']``.

    Notes
    -----
    This is a wrapper around two functions of the same name,
    which generate ``X`` in the signomial and polynomial cases.

    The signomial and polynomial cases are implemented in
    ``sageopt.relaxations.sage_sigs.sage_feasibility`` and
    ``sageopt.relaxations.sage_polys.sage_feasibility`` respectively.

    Refer to those functions for detailed documentation.

    """
    if isinstance(f, Polynomial):
        X = _poly_cond_sage_data(f, gts, eqs)
        return X
    elif isinstance(f, Signomial):
        X = _sig_cond_sage_data(f, gts, eqs)
        return X
    else:
        raise ValueError('"f" must be a Signomial or Polynomial.')



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""


class Constraint(object):

    def is_affine(self):
        raise NotImplementedError()

    def is_elementwise(self):
        raise NotImplementedError()

    def is_setmem(self):
        raise NotImplementedError()

    def variables(self):
        raise NotImplementedError()

    def conic_form(self):
        """
        :return: A_vals, A_rows, A_cols, b, K, sep_K
            A_vals - list (of floats)
            A_rows - numpy 1darray (of integers)
            A_cols - list (of integers)
            b - numpy 1darray (of floats)
            K - list (of coniclifts Cone objects)

        """
        raise NotImplementedError()

    def violation(self):
        raise NotImplementedError()



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import sageopt.coniclifts as cl
from sageopt.symbolic.signomials import Signomial
from sageopt.relaxations.symbolic_correspondences import moment_reduction_array
import numpy as np
from scipy.optimize import fmin_cobyla


def is_feasible(x, greater_than_zero, equal_zero, ineq_tol=1e-8, eq_tol=1e-8):
    if any([g(x) < -ineq_tol for g in greater_than_zero]):
        return False
    if any([abs(g(x)) > eq_tol for g in equal_zero]):
        return False
    return True


def local_refine(f, gts, eqs, x0, rhobeg=1, rhoend=1e-7, maxfun=10000):
    res = fmin_cobyla(f, x0, gts + eqs + [-g for g in eqs],
                      rhobeg=rhobeg, rhoend=rhoend, maxfun=maxfun)
    return res


def sig_solrec(prob, ineq_tol=1e-8, eq_tol=1e-6, skip_ls=False):
    con = prob.user_cons[0]
    if not con.name == 'Lagrangian SAGE dual constraint':
        raise RuntimeError('Unexpected first constraint in dual SAGE relaxation.')
    f = prob.associated_data['f']
    # Recover any constraints present in "prob"
    lag_gts, lag_eqs = [], []
    if 'gts' in prob.associated_data:
        # only happens in "constrained_sage_dual".
        lag_gts = prob.associated_data['gts']
        lag_eqs = prob.associated_data['eqs']
    lagrangian = _make_dummy_lagrangian(f, lag_gts, lag_eqs)
    gts = lag_gts + prob.associated_data['X']['gts']
    eqs = lag_eqs + prob.associated_data['X']['eqs']
    # Search for solutions which meet the feasibility criteria
    v = con.v.value
    if np.any(np.isnan(v)):
        return None
    if not hasattr(con, 'alpha'):
        alpha = con.lifted_alpha[:, :con.n]
    else:
        alpha = con.alpha
    dummy_modulated_lagrangian = Signomial(alpha, np.ones(shape=(alpha.shape[0],)))
    alpha_reduced = lagrangian.alpha
    modulator = prob.associated_data['modulator']
    M = moment_reduction_array(lagrangian, modulator, dummy_modulated_lagrangian)
    if skip_ls:
        mus0 = []
    else:
        mus0 = _least_squares_solution_recovery(alpha_reduced, con, v, M, gts, eqs, ineq_tol, eq_tol)
    mus1 = _dual_age_cone_solution_recovery(con, v, M, gts, eqs, ineq_tol, eq_tol)
    mus = mus0 + mus1
    mus.sort(key=lambda mu: f(mu))
    return mus


def _least_squares_solution_recovery(alpha_reduced, con, v, M, gts, eqs, ineq_tol, eq_tol):
    v_reduced = M @ v
    log_v_reduced = np.log(v_reduced)
    if isinstance(con, cl.DualCondSageCone):
        mu_ls = _constrained_least_squares(con, alpha_reduced, log_v_reduced)
    else:
        try:
            mu_ls = np.linalg.lstsq(alpha_reduced, log_v_reduced, rcond=None)[0]
        except np.linalg.linalg.LinAlgError:
            mu_ls = None
    if mu_ls is not None and is_feasible(mu_ls, gts, eqs, ineq_tol, eq_tol):
        return [mu_ls]
    else:
        return []


def _constrained_least_squares(con, alpha, log_v):
    A, b, K = con.A, con.b, con.K
    A = np.asarray(A)
    x = cl.Variable(shape=(A.shape[1],))
    t = cl.Variable(shape=(1,))
    cons = [cl.vector2norm(log_v - alpha @ x[:con.n]) <= t,
            cl.PrimalProductCone(A @ x + b, K)]
    prob = cl.Problem(cl.MIN, t, cons)
    cl.clear_variable_indices()
    res = prob.solve(verbose=False)
    if res[0] == cl.SOLVED:
        mu_ls = x.value[:con.n]
        return mu_ls
    else:
        return None


def _dual_age_cone_solution_recovery(con, v, M, gts, eqs, ineq_tol, eq_tol):
    mus_exist = list(con.mu_vars.keys())
    # build a matrix whose columns are simple candidate solutions to an optimization problem.
    raw_xs = []
    for i in mus_exist:
        mu_i = con.mu_vars[i]
        xi = (mu_i.value / v[i]).reshape((-1, 1))
        raw_xs.append(xi)
    raw_xs = np.hstack(raw_xs)
    # build a matrix "weights", whose rows specify convex combination coefficients.
    M_interest = M[:, mus_exist]
    v_interest = v[mus_exist].ravel()
    v_reduced = M_interest @ v_interest
    keep_rows = v_reduced != 0
    M_interest = M_interest[keep_rows, :]
    v_reduced = v_reduced[keep_rows]
    weights = M_interest / v_reduced[:, None]
    weights = (weights.T * v_interest[:, None]).T
    # the rows of "weights" now define convex combination vectors.
    reduced_xs = raw_xs @ weights.T
    # concatenate the new solutions (reduced_xs) with original solutions (raw_xs)
    all_xs = np.hstack((raw_xs, reduced_xs))
    all_xs = np.unique(all_xs, axis=1)
    # check each candidate solution for feasibility (up to desired tolerances)
    # and return the result
    mus = []
    for xi in all_xs.T:
        if is_feasible(xi, gts, eqs, ineq_tol, eq_tol):
            mus.append(xi)
    return mus


def _satisfies_AbK_constraints(A, b, K, mu, ineq_tol):
    # This function cannot be trusted when some component mu[i] == -np.inf.
    if np.any(np.isnan(mu)):
        return False
    x = cl.Variable(shape=(A.shape[1],))
    t = cl.Variable(shape=(1,))
    mu_flat = mu.ravel()
    where_finite = np.where(mu > -np.inf)[0]
    cons = [cl.vector2norm(mu_flat[where_finite] - x[where_finite]) <= t, cl.PrimalProductCone(A @ x + b, K)]
    prob = cl.Problem(cl.MIN, t, cons)
    cl.clear_variable_indices()
    res = prob.solve(verbose=False)
    if res[0] == cl.SOLVED and res[1] <= ineq_tol + 1e-8:
        return True
    else:
        return False


def _make_dummy_lagrangian(f, gts, eqs):
    dummy_gamma = cl.Variable(shape=())
    dummy_slacks = cl.Variable(shape=(len(gts),))
    dummy_multipliers = cl.Variable(shape=(len(eqs)))
    ineq_term = sum([gts[i] * dummy_slacks[i] for i in range(len(gts))])
    eq_term = sum([eqs[i] * dummy_multipliers[i] for i in range(len(eqs))])
    dummy_L = f - dummy_gamma - ineq_term - eq_term
    cl.clear_variable_indices()
    return dummy_L



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from sageopt.coniclifts.constraints.constraint import Constraint


class SetMembership(Constraint):

    def is_affine(self):
        return True

    def is_setmem(self):
        return True

    def is_elementwise(self):
        return False

    def variables(self):
        raise NotImplementedError()

    def conic_form(self):
        raise NotImplementedError()



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
from sageopt.coniclifts.constraints.set_membership.setmem import SetMembership
from sageopt.coniclifts.cones import Cone


class PSD(SetMembership):

    _CURVATURE_CHECK_ = False

    _SYMMETRY_CHECK_ = True

    _CONSTRAINT_ID_ = 0

    _LMI_OPS_ = ('<<', '>>')

    def __init__(self, arg):
        self.id = PSD._CONSTRAINT_ID_
        PSD._CONSTRAINT_ID_ += 1
        self.arg = arg
        self.expr = None
        if PSD._SYMMETRY_CHECK_:
            from sageopt.coniclifts.base import Expression
            expr_sym = (arg + arg.T) / 2
            if not Expression.are_equivalent(arg, expr_sym):
                raise RuntimeError('Argument to LMI was not symmetric.')
        pass

    def variables(self):
        return self.arg.variables()

    def conic_form(self):
        from sageopt.coniclifts.base import Expression, ScalarVariable
        expr = np.triu(self.arg).view(Expression)
        expr = expr[np.triu_indices(expr.shape[0])]
        K = [Cone('P', expr.size)]
        b = np.empty(shape=(expr.size,))
        A_rows, A_cols, A_vals = [], [], []
        for i, se in enumerate(expr):
            b[i] = se.offset
            if len(se.atoms_to_coeffs) == 0:
                A_rows.append(i)
                A_cols.append(ScalarVariable.curr_variable_count())
                A_vals.append(0)  # make sure scipy infers correct dimensions later on.
            else:
                A_rows += [i] * len(se.atoms_to_coeffs)
                cols_and_coeff = [(a.id, c) for a, c in se.atoms_to_coeffs.items()]
                A_cols += [atom_id for (atom_id, _) in cols_and_coeff]
                A_vals += [c for (_, c) in cols_and_coeff]
        return [(A_vals, np.array(A_rows), A_cols, b, K)]



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
import scipy.sparse as sp
from sageopt.coniclifts.reformulators import separate_cone_constraints, build_cone_type_selectors
from sageopt.coniclifts.problems.solvers.solver import Solver
from sageopt.coniclifts.standards import constants as CL_CONSTANTS
from sageopt.coniclifts.cones import Cone
import copy


class Mosek(Solver):

    _SDP_SUPPORT_ = False

    _CO_TOL_NEAR_REL_ = 1000

    @staticmethod
    def apply(c, A, b, K, destructive, compilation_options):
        # This function is analogous to "apply(...)" in cvxpy's mosek_conif.py.
        #
        # Main obstacle: even after running (A,b,K) through "separate_cone_constraints", the PSD constraints are
        # still problematic. Reason being, MOSEK doesn't work with vectorized semidefinite variables. It assumes
        # that these are declared separately from the vectorized variable, and the contribution of semidefinite
        # variables to linear [in]equalities must be specified separately from the contributions of the vectorized
        # variable.
        #
        # Plan: call "separate_cone_constraints(...)" with dont_sep={'0','+','P'}, and then handle the semidefinite
        # constraints just like I do in CVXPY. For CVXPY I sort the cone types in (A,b,K) so that linear matrix
        # inequalities come last. Then for every LMI I declare a slack variable.
        # I would then essentially copy the flow of
        # https://github.com/cvxgrp/cvxpy/blob/16ad9adad944d4cb34275e2f7dd0e27788b47b18/ ...
        #       .... cvxpy/reductions/solvers/conic_solvers/mosek_conif.py#L389
        # in order to enforce proper equality constraints on these slack variables.
        #
        if not Mosek._SDP_SUPPORT_:
            if any([co.type == 'P' for co in K]):
                raise NotImplementedError('This functionality is being put on hold.')
        if not destructive:
            A = A.copy()
            b = b.copy()
            K = copy.deepcopy(K)
            c = c.copy()
        if 'avoid_slacks' in compilation_options:
            avoid_slacks = compilation_options['avoid_slacks']
        else:
            avoid_slacks = False
        A, b, K, sep_K, scale, trans = separate_cone_constraints(A, b, K,
                                                                 destructive=True,
                                                                 dont_sep={'0', '+', 'P'},
                                                                 avoid_slacks=avoid_slacks)
        # A,b,K now reflect a conic system where "x" has been replaced by "np.diag(scale) @ (x + trans)"
        c = np.hstack([c, np.zeros(shape=(A.shape[1] - len(c)))])
        objective_offset = c @ trans
        c = c * scale
        type_selectors = build_cone_type_selectors(K)
        # Below: inequality constraints that the "user" intended to give to MOSEK.
        A_ineq = A[type_selectors['+'], :]
        b_ineq = b[type_selectors['+']]
        # Below: equality constraints that the "user" intended to give to MOSEK
        A_z = A[type_selectors['0'], :]
        b_z = b[type_selectors['0']]
        # Finally: the matrix "A" and vector "u_c" that appear in the MOSEK documentation as the standard form for
        # an SDP that includes vectorized variables. We use "b" instead of "u_c". It's value in the MOSEK Task will
        # later be specified with MOSEK's "putconboundlist" function.
        A = -sp.vstack([A_ineq, A_z], format='csc')
        b = np.hstack([b_ineq, b_z])
        K = [Cone('+', A_ineq.shape[0]), Cone('0', A_z.shape[0])]
        # Return values
        data = {'A': A, 'b': b, 'K': K, 'sep_K': sep_K, 'c': c}
        inv_data = {'scaling': scale, 'translation': trans, 'objective_offset': objective_offset, 'n': A.shape[1]}
        return data, inv_data

    @staticmethod
    def solve_via_data(data, params):
        # This function is analogous to "solve_via_data(...)" in cvxpy's mosek_conif.py.
        import mosek
        env = mosek.Env()
        task = env.Task(0, 0)
        if params['verbose']:
            Mosek.set_verbosity_params(env, task)
        Mosek.set_default_solver_settings(task)

        # The following lines recover problem parameters, and define helper constants.
        if not params['destructive']:
            data['A'] = data['A'].copy()
            data['b'] = data['b'].copy()
            data['K'] = data['K'].copy()
        c, A, b, K, sep_K = data['c'], data['A'], data['b'], data['K'], data['sep_K']
        n = A.shape[1]
        m = A.shape[0]

        # Define variables and cone constraints
        task.appendvars(n)
        task.putvarboundlist(np.arange(n, dtype=int),
                             [mosek.boundkey.fr] * n, np.zeros(n), np.zeros(n))
        for co in sep_K:
            if co.type == 'S':
                indices = co.annotations['col mapping']
                task.appendcone(mosek.conetype.quad, 0.0, indices)
            elif co.type == 'e':
                co_cols = co.annotations['col mapping']
                indices = [co_cols[1], co_cols[2], co_cols[0]]
                task.appendcone(mosek.conetype.pexp, 0.0, indices)
            else:
                raise RuntimeError('Unknown separated cone ' + str(co[0]) + '.')

        # Define linear inequality and equality constraints.
        task.appendcons(m)
        rows, cols, vals = sp.find(A)
        task.putaijlist(rows.tolist(), cols.tolist(), vals.tolist())
        type_constraint = [mosek.boundkey.up] * K[0].len + [mosek.boundkey.fx] * K[1].len
        task.putconboundlist(np.arange(m, dtype=int), type_constraint, b, b)

        # Define the objective
        task.putclist(np.arange(len(c)), c)
        task.putobjsense(mosek.objsense.minimize)

        # Optimize the Mosek Task and return the result.
        task.optimize()
        if params['verbose']:
            task.solutionsummary(mosek.streamtype.msg)
        return {'env': env, 'task': task, 'params': params}

    @staticmethod
    def parse_result(solver_output, inv_data, var_mapping):
        """
        :param solver_output: a dictionary containing the mosek Task and Environment that
        were constructed at solve-time. Also contains any parameters that were passed at solved-time.

        :param inv_data: a dictionary with keys 'scaling' and 'translation'. The primal variable
        values returned by Mosek should be ...
            (1) scaled by the inverse of inv_data['scaling'], and
            (2) translated by the negative of inv_data['translation'].
        Then the objective value should be reduced by inv_data['objective_offset'].

        :param var_mapping: a dictionary mapping names of coniclifts Variables to arrays
        of indices. The array var_mapping['my_var'] contains the indices in
        "solver_output['x']" of the coniclifts Variable object named 'my_var'.

        :return: problem_status, user_variable_values, problem_value. The first of these is a string
        (coniclifts.solved, coniclifts.inaccurate, or coniclifts.failed). The second is a dictionary
        from coniclifts Variable names to numpy arrays containing the values of those Variables at
        the returned solution. The last of these is either a real number, or +/-np.Inf, or np.NaN.
        """
        import mosek
        task = solver_output['task']
        sol = mosek.soltype.itr
        solution_status = task.getsolsta(sol)
        variable_values = dict()
        if solution_status == mosek.solsta.optimal:
            # optimal
            problem_status = CL_CONSTANTS.solved
            problem_value = task.getprimalobj(sol) - inv_data['objective_offset']
            x0 = [0.] * inv_data['n']
            task.getxxslice(sol, 0, len(x0), x0)
            x0 = np.array(x0)
            variable_values = Mosek.load_variable_values(x0, inv_data, var_mapping)
        elif solution_status == mosek.solsta.dual_infeas_cer:
            # unbounded
            problem_status = CL_CONSTANTS.solved
            problem_value = -np.Inf
        elif solution_status == mosek.solsta.prim_infeas_cer:
            # infeasible
            problem_status = CL_CONSTANTS.solved
            problem_value = np.Inf
        else:
            # some kind of solver failure.
            problem_status = CL_CONSTANTS.failed
            variable_values = dict()
            problem_value = np.NaN
        return problem_status, variable_values, problem_value

    @staticmethod
    def is_installed():
        try:
            import mosek
            return True
        except ImportError:
            return False

    @staticmethod
    def set_verbosity_params(env, task):
        # If verbose, then set default logging parameters.
        import mosek
        import sys

        def stream_printer(text):
            sys.stdout.write(text)
            sys.stdout.flush()
        print('\n')
        env.set_Stream(mosek.streamtype.log, stream_printer)
        task.set_Stream(mosek.streamtype.log, stream_printer)
        task.putintparam(mosek.iparam.log_presolve, 2)

    # noinspection SpellCheckingInspection
    @staticmethod
    def set_default_solver_settings(task):
        import mosek
        task.putintparam(mosek.iparam.num_threads, 0)
        task.putdouparam(mosek.dparam.intpnt_co_tol_near_rel, Mosek._CO_TOL_NEAR_REL_)
        # pdsafe = 1e1
        # task.putdouparam(mosek.dparam.intpnt_tol_psafe, pdsafe)
        # task.putdouparam(mosek.dparam.intpnt_tol_dsafe, pdsafe)
        # task.putdouparam(mosek.dparam.intpnt_co_tol_infeas, 1e-8)
        # task.putdouparam(mosek.dparam.intpnt_tol_path, 1.0e-1)
        # task.putintparam(mosek.iparam.intpnt_scaling, mosek.scalingtype.free)
        # task.putintparam(mosek.iparam.intpnt_solve_form, mosek.solveform.primal)
        # task.putintparam(mosek.iparam.intpnt_max_num_cor, int(1e8))
        # task.putintparam(mosek.iparam.intpnt_max_num_refinement_steps, int(1e8))
        # task.putintparam(mosek.iparam.intpnt_regularization_use, mosek.onoffkey.on)
        # task.putdouparam(mosek.dparam.intpnt_co_tol_rel_gap, 1.0)
        # task.putdouparam(mosek.dparam.intpnt_co_tol_mu_red, 1.0)
        # task.putintparam(mosek.iparam.intpnt_off_col_trh, 5)
        # task.putintparam(mosek.iparam.presolve_use, mosek.presolvemode.on)
        # task.putintparam(mosek.iparam.presolve_eliminator_max_num_tries, 10)
        # task.putintparam(mosek.iparam.presolve_lindep_use, mosek.onoffkey.on)
        # task.putintparam(mosek.iparam.presolve_lindep_abs_work_trh, int(1e8))
        # task.putintparam(mosek.iparam.presolve_lindep_rel_work_trh, int(1e8))
        # task.putintparam(mosek.iparam.presolve_eliminator_max_fill, int(1e8))
        # task.putintparam(mosek.iparam.log_response, 3)
        # task.putintparam(mosek.iparam.log_feas_repair, 3)
        pass

    @staticmethod
    def load_variable_values(x0, inv_data, var_mapping):
        x = np.power(inv_data['scaling'], -1) * x0 - inv_data['translation']
        var_values = dict()
        for var_name in var_mapping:
            var_values[var_name] = x[var_mapping[var_name]]
        return var_values



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import sageopt.coniclifts as cl
from sageopt.symbolic.polynomials import Polynomial, standard_poly_monomials
from sageopt.relaxations.symbolic_correspondences import moment_reduction_array
from sageopt.relaxations.sig_solution_recovery import is_feasible
import numpy as np
from scipy.optimize import fmin_cobyla
import itertools


def local_refine_polys_from_sigs(f, gts, eqs, x0, rhobeg=1.0, rhoend=1e-7, maxfun=10000):
    """
    When faced with a polynomial optimization problem over nonnegative variables, one should
    formulate the problem in terms of signomials. This reformulation is without loss of generality
    from the perspective of solving a SAGE relaxation, but the local-refinement stage of solution
    recovery is somewhat different.

    This is a helper function which ...
        (1) accepts signomial problem data (representative of a desired polynomial optimization problem),
        (2) transforms the signomial data into equivalent polynomial data, and
        (3) performs local refinement on the polynomial data, via the COBYLA solver.

    This function may be important if a polynomial optimization problem has some variable equal to zero
    in an optimal solution.

    :param f: a Signomial, defining the objective function to be minimized. From "f" we will construct
    a polynomial "p" where p(y) = f(log(y)) for all elementwise positive vectors y.

    :param gts: a list of Signomials, each defining an inequality constraint g(x) >= 0.
    From this list, we will construct a list of polynomials gts_poly, so that every g0 in gts has a
    polynomial representative g1 in gts_poly, satisfying g1(y) = g0(log(y)) for all elementwise
    positive vectors y.

    :param eqs: a list of Signomials, each defining an equality constraint g(x) == 0.
    From this list, we will construct a list of polynomials gts_poly, so that every g0 in eqs has a
    polynomial representative g1 in eqs_poly, satisfying g1(y) = g0(log(y)) for all elementwise
    positive vectors y.

    :param x0: an initial condition for the signomial optimization problem
        min{ f(x) |  g(x) >= 0 for g in gts, g(x) == 0 for g in eqs }.

    :param rhobeg: COBYLA's initial radius of the search space around y0 = exp(x0).

    :param rhoend: COBYLA terminates once the radius of the search space reaches this value.

    :param maxfun: The maximum number of iterations used by the COBYLA solver.

    :return: The output of COBYLA for the polynomial optimization problem
        min{ p(y) | g(y) >= 0 for g in gts_poly, g(y) == 0 for g in eqs_poly, y >= 0 }
    with initial condition y0 = exp(x0).
    """
    x0 = np.exp(x0)
    gts = [g.as_polynomial() for g in gts]
    x = standard_poly_monomials(x0.size)
    gts += [x[i] for i in range(x0.size)]  # Decision variables must be nonnegative.
    eqs = [g.as_polynomial() for g in eqs]
    f = f.as_polynomial()
    res = fmin_cobyla(f, x0, gts + eqs + [-g for g in eqs],
                      rhobeg=rhobeg, rhoend=rhoend, maxfun=maxfun)
    return res


def poly_solrec(prob, ineq_tol=1e-8, eq_tol=1e-6, zero_tol=1e-20, hueristic=False, all_signs=True, skip_ls=False):
    # implemented only for poly_constrained_dual (not yet implemented for sage_poly_dual).
    f = prob.associated_data['f']
    lag_gts = prob.associated_data['gts']
    lag_eqs = prob.associated_data['eqs']
    lagrangian = _make_dummy_lagrangian(f, lag_gts, lag_eqs)
    con = prob.user_cons[0]
    if isinstance(con, cl.DualCondSageCone):
        alpha = con.lifted_alpha[:, :con.n]
    else:
        alpha = con.alpha
    dummy_modulated_lagrangian = Polynomial(alpha, np.ones(shape=(alpha.shape[0],)))  # coefficients dont matter
    modulator = prob.associated_data['modulator']
    v = prob.associated_data['v_poly'].value  # possible that v_sig and v are the same
    if np.any(np.isnan(v)):
        return []
    M = moment_reduction_array(lagrangian, modulator, dummy_modulated_lagrangian)
    v_reduced = M @ v
    alpha_reduced = lagrangian.alpha
    mags = variable_magnitudes(con, alpha_reduced, v_reduced, zero_tol, skip_ls)
    signs = variable_sign_patterns(alpha_reduced, v_reduced, hueristic, all_signs)
    # Now we need to build the candidate solutions, and check them for feasibility.
    gts = lag_gts + [g for g in prob.associated_data['X']['gts']]
    eqs = lag_eqs + [g for g in prob.associated_data['X']['eqs']]
    solutions = []
    for mag in mags:
        for sign in signs:
            x = mag * sign  # elementwise
            if is_feasible(x, gts, eqs, ineq_tol, eq_tol):
                solutions.append(x)
    solutions.sort(key=lambda xi: f(xi))
    return solutions


def variable_magnitudes(con, alpha_reduced, v_reduced, zero_tol, skip_ls):
    # This is essentially "Algorithm 3" in the conditional SAGE paper.
    v_sig = con.v.value
    M_sig = np.eye(v_sig.size)
    mags0 = _dual_age_cone_magnitude_recovery(con, v_sig, M_sig)
    if skip_ls:
        return mags0
    else:
        abs_mom_mag = _least_squares_magnitude_recovery(con, alpha_reduced, v_reduced, zero_tol)
        if abs_mom_mag is None:
            mags1 = []
        else:
            mags1 = [abs_mom_mag]
        mags = mags0 + mags1
        return mags


def _dual_age_cone_magnitude_recovery(con, v_sig, M_sig):
    mus_exist = list(con.mu_vars.keys())
    # build a matrix whose columns are simple candidate solutions to an optimization problem.
    raw_ys = []
    for i in mus_exist:
        mu_i = con.mu_vars[i]
        yi = (mu_i.value / v_sig[i]).reshape((-1, 1))
        raw_ys.append(yi)
    raw_ys = np.hstack(raw_ys)
    # build a matrix "weights", whose rows specify convex combination coefficients.
    M_interest = M_sig[:, mus_exist]
    v_interest = v_sig[mus_exist].ravel()
    v_reduced = M_interest @ v_interest
    keep_rows = v_reduced != 0
    M_interest = M_interest[keep_rows, :]
    v_reduced = v_reduced[keep_rows]
    weights = M_interest / v_reduced[:, None]
    weights = (weights.T * v_interest[:, None]).T
    # the rows of "weights" now define convex combination vectors.
    reduced_ys = raw_ys @ weights.T
    # concatenate the new solutions (reduced_ys) with original solutions (raw_ys)
    all_ys = np.hstack((raw_ys, reduced_ys))
    all_ys = np.unique(all_ys, axis=1).astype(np.float128)
    all_xs = np.exp(all_ys)
    mus = [xi for xi in all_xs.T]
    return mus


def _least_squares_magnitude_recovery(con, alpha_reduced, v_reduced, zero_tol):
    v_abs = np.abs(v_reduced).ravel()
    if isinstance(con, cl.DualCondSageCone):
        n = con.lifted_n
    else:
        n = con.n
    if n > con.n:
        padding = np.zeros(shape=(alpha_reduced.shape[0], n - con.n))
        alpha_reduced = np.hstack((alpha_reduced, padding))
    y = cl.Variable(shape=(n,), name='abs moment mag recovery')
    are_nonzero = v_abs > np.sqrt(zero_tol)
    t = cl.Variable(shape=(1,), name='t')
    residual = alpha_reduced[are_nonzero, :] @ y - np.log(v_abs[are_nonzero])
    constraints = [cl.vector2norm(residual) <= t]
    if np.any(~are_nonzero):
        tempcon = alpha_reduced[~are_nonzero, :] @ y <= np.log(zero_tol)
        constraints.append(tempcon)
    if isinstance(con, cl.DualCondSageCone):
        A, b, K = np.asarray(con.A), con.b, con.K
        tempcon = cl.PrimalProductCone(A @ y + b, K)
        constraints.append(tempcon)
    prob = cl.Problem(cl.MIN, t, constraints)
    prob.solve(verbose=False)
    cl.clear_variable_indices()
    if prob.status == cl.SOLVED and prob.value < np.inf:
        mag = np.exp(y.value.astype(np.float128))
        return mag
    else:
        return None


def variable_sign_patterns(alpha, moments, hueristic=False, all_signs=True):
    # This is essentially "Algorithm 4" in the conditional SAGE paper.
    #
    # ignore signs of variables which only participate in even monomials.
    #
    # variables that participate in monomials "j" with moments[j] == 0
    # are given nonnegative signs.
    m, n = alpha.shape
    x0, alpha1, U, W = linear_system_negatives(alpha, moments)
    if x0 is None:
        if not hueristic:
            return []
        else:
            x0 = greedy_weighted_cut_negatives(alpha, moments)
            y0 = np.ones(n,)
            y0[x0 == 1] = -1
            return [y0]
    elif alpha1 is None:
        y0 = np.ones(n,)
        return [y0]
    else:
        if all_signs:
            arref, p = mod2rref(alpha1)
            N0 = mod2nullspace(arref, p)
        else:
            N0 = [np.zeros(alpha1.shape[1],)]
        signs = []
        for vec0 in N0:
            vec = np.zeros(n,)
            vec[W] = vec0
            vec = np.mod(vec + x0, 2).astype(int)
            y = np.ones(n,)
            y[vec == 1] = -1
            signs.append(y)
        return signs


def linear_system_negatives(alpha, moments):
    """
    :param alpha: an m-by-n array of integers.
    :param moments: a vector of length m.
    :return: a length-n vector with entries from {0, 1}. 1 means "negative",
    "0" means "nonnegative".
    """
    m, n = alpha.shape
    alpha = np.mod(alpha, 2).astype(int)
    U = [i for i in range(m) if abs(moments[i]) > 0 and np.any(alpha[i, :] > 0)]
    # If moment[i] == 0, then there is no sign we need to be consistent with.
    # If alpha[i,:] == 0 (after taking alpha mod 2), then we must have moments[i] >= 0,
    # and the resulting row of the linear system is trivially satisfied. Therefore
    # we restrict our attention to rows "i" with some alpha[i, j] = 1 (mod 2) for some
    # j, without loss of generality.
    if len(U) == 0:
        return np.zeros(n,), None, None, None
    W = [j for j in range(n) if np.any(alpha[U, j] > 0)]
    # If, out of the remaining rows in the linear system, the variable at coordinate "j"
    # only participates in even monomials, then we remove these columns from the matrix
    # "alpha". Removing these columns has a very modest speed improvement here (when
    # finding a single solution to a linear system), but can have a very large effect
    # when enumerating all vectors in the mod-2 nullspace of the smaller matrix.
    if len(W) == 0:
        return np.zeros(n,), None, None, None
    alpha = alpha[U, :]
    alpha = alpha[:, W]
    b = ((moments < 0)[U]).astype(int)
    x_W = mod2linsolve(alpha, b)
    if x_W is None:
        return None, alpha, U, W
    else:
        x = np.zeros(n,)
        x[W] = x_W
        return x, alpha, U, W


def _lin_indep_subset_negatives(alpha, b):
    _, lin_indep_row_idxs = mod2rref(alpha.T)
    alpha = alpha[lin_indep_row_idxs, :]
    b = b[lin_indep_row_idxs]
    x = mod2linsolve(alpha, b)
    return x


def greedy_weighted_cut_negatives(alpha, v):
    m, n = alpha.shape
    y = np.ones(n,)
    v = v.ravel()
    coordinates_remaining = set(range(n))
    merit_of_switch = np.zeros(n,)
    for iter_num in range(n):
        for i in coordinates_remaining:
            before = np.dot(v, np.prod(np.power(y, alpha), axis=1))
            y[i] = -1
            after = np.dot(v, np.prod(np.power(y, alpha), axis=1))
            y[i] = 1
            merit_of_switch[i] = after - before
        best_candidate = np.argmax(merit_of_switch)
        best_candidate = best_candidate[(0,) * best_candidate.ndim]
        if merit_of_switch[best_candidate] > 0:
            # if "merit" is positive, then the vector y^alpha
            # points "more" along the direction of "v" than it
            # did before. Ideally we want y^alpha to point exactly
            # along sign(v), but the intended use-case of this function
            # is when no such value of y in {+1, 1}^n can have y^alpha
            # match sign(v). Therefore it makes more sense to use a
            # weighted measure of similarity -- taking the inner-product
            # of v and y^alpha is a reasonable measure of similarity.
            y[best_candidate] = -1
            merit_of_switch[best_candidate] = -1
            coordinates_remaining.remove(best_candidate)
        else:
            break
    x = y.copy()
    x[y == -1] = 1
    x[y == 1] = 0
    return x


def mod2rref(A, forward_only=False):
    """
    :param A: an integer array with A.shape == (m, n).
    :param forward_only: boolean. False means "return reduced row-echelon form", True means
    "return row echelon form".
    :return: the (reduced) row-echelon form of A (in mod 2 arithmetic) and corresponding pivot columns
    """
    A = np.mod(A, 2).astype(int)
    h, k = 0, 0
    m, n = A.shape
    pivot_columns = []
    # Forward elimination
    while h < m and k < n:
        i_max = h + np.argmax(A[h:, k])
        if A[i_max, k] == 0:
            k += 1
        else:
            # swap rows "h" and "i_max"
            pivot_columns.append(k)
            row_h = A[h, :].copy()
            row_imax = A[i_max, :].copy()
            A[h, :] = row_imax
            A[i_max, :] = row_h
            for i in range(h+1, m):
                if A[i, k] > 0:
                    A[i, :] = np.mod(A[i, :] - A[h, :], 2)
            h += 1
            k += 1
    # Back substitution
    if not forward_only:
        for pr, pc in enumerate(pivot_columns):
            for row in range(pr-1, -1, -1):
                if A[row, pc] > 0:
                    # ^ That isn't necessarily valid when a matrix is wide, and contains
                    # a pivot column far down the list of columns.
                    A[row, pc:] = np.mod(A[row, pc:] - A[pr, pc:], 2)
    return A, pivot_columns


def mod2nullspace_basis(arref, p):
    m, n = arref.shape
    r = len(p)
    F = {j for j in range(n) if j not in p}  # set of free variables
    N = np.zeros(shape=(n, len(F)), dtype=int)
    for ell, f in enumerate(F):
        N[f, ell] = 1
        for i in range(r):
            N[p[i], ell] = arref[i, f]
    return N


def mod2nullspace(arref, p):
    N = mod2nullspace_basis(arref, p)
    N = [vec for vec in N.T]
    nullspace = {(0,) * arref.shape[1]}
    powerset_iterator = itertools.chain.from_iterable(itertools.combinations(N, r) for r in range(1, len(N)+1))
    for vecs in powerset_iterator:
        vec = np.mod(np.sum(vecs, axis=0), 2).astype(int)
        vec = tuple(vec.tolist())
        nullspace.add(vec)
    N = [np.array(vec) for vec in nullspace]
    return N


def mod2linsolve(A, b):
    """
    :param A: m-by-n
    :param b: m
    :return: solution to A @ x == b (mod 2), or None.
    """
    m, n = A.shape
    A0 = np.hstack([A, b.reshape((-1, 1))])
    A1, pivcols = mod2rref(A0, forward_only=True)
    augmented_rank = len(pivcols)
    if augmented_rank > 0 and pivcols[-1] == n:
        return None
    else:
        b1 = A1[:n, A.shape[1]]
        A1 = A1[:n, :A.shape[1]]
        x = np.zeros(shape=(n,), dtype=int)
        row = len(pivcols) - 1
        for pc in reversed(pivcols):
            x[pc] = (b1[row] - np.dot(A1[row, (pc+1):], x[(pc+1):])) % 2
            row -= 1
        return x


def _make_dummy_lagrangian(f, gts, eqs):
    dummy_gamma = cl.Variable(shape=())
    dummy_slacks = cl.Variable(shape=(len(gts),))
    dummy_multipliers = cl.Variable(shape=(len(eqs)))
    ineq_term = sum([gts[i] * dummy_slacks[i] for i in range(len(gts))])
    eq_term = sum([eqs[i] * dummy_multipliers[i] for i in range(len(eqs))])
    dummy_L = f - dummy_gamma - ineq_term - eq_term
    cl.clear_variable_indices()
    return dummy_L



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
import scipy.special as special_functions
from sageopt.coniclifts.base import NonlinearScalarAtom, Expression, ScalarExpression, Variable
from sageopt.coniclifts.utilities import array_index_iterator
from sageopt.coniclifts.cones import Cone


def relent(x, y, elementwise=False):
    if not isinstance(x, Expression):
        x = Expression(x)
    if not isinstance(y, Expression):
        y = Expression(y)
    if x.size != y.size:
        raise RuntimeError('Incompatible arguments to relent.')
    if elementwise:
        expr = np.empty(shape=x.shape, dtype=object)
        for tup in array_index_iterator(expr.shape):
            expr[tup] = ScalarExpression({RelEnt(x[tup], y[tup]): 1}, 0)
        return expr.view(Expression)
    else:
        x = x.ravel()
        y = y.ravel()
        d = dict((RelEnt(x[i], y[i]), 1) for i in range(x.size))
        return ScalarExpression(d, 0).as_expr()


class RelEnt(NonlinearScalarAtom):

    _REL_ENT_COUNTER_ = 0

    @staticmethod
    def __atom_text__():
        return 'RelEnt'

    def __init__(self, x, y):
        """
        Used to represent the epigraph of  "x * ln(x / y)".

        https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.rel_entr.html#scipy.special.rel_entr
        """
        self._args = (self.parse_arg(x), self.parse_arg(y))
        self._id = RelEnt._REL_ENT_COUNTER_
        RelEnt._REL_ENT_COUNTER_ += 1
        v = Variable(shape=(), name='_rel_ent_epi[' + str(self.id) + ']_')
        self._epigraph_variable = v[()].scalar_variables()[0]

    def is_convex(self):
        return True

    def is_concave(self):
        return False

    def epigraph_conic_form(self):
        """
        Generate conic constraint for epigraph
            self.args[0] * ln( self.args[0] / self.args[1] ) <= self._epigraph_variable.

        :return:
        """
        b = np.zeros(3,)
        K = [Cone('e', 3)]
        # ^ initializations
        A_rows, A_cols, A_vals = [0], [self._epigraph_variable.id], [-1]
        # ^ first row
        x = self.args[0]
        A_rows += (len(x)-1) * [2]
        A_cols += [var.id for var, co in x[:-1]]
        A_vals += [co for var, co in x[:-1]]
        b[2] = x[-1][1]
        # ^ third row
        y = self.args[1]
        A_rows += (len(y)-1) * [1]
        A_cols += [var.id for var, co in y[:-1]]
        A_vals += [co for var, co in y[:-1]]
        b[1] = y[-1][1]
        # ^ second row
        return A_vals, np.array(A_rows), A_cols, b, K

    def value(self):
        vals = []
        for i in [0, 1]:
            arg_as_list = self.args[i]
            d = dict(arg_as_list[:-1])
            arg_se = ScalarExpression(d, arg_as_list[-1][1], verify=False)
            arg_val = arg_se.value()
            vals.append(arg_val)
        val = special_functions.rel_entr(vals[0], vals[1])
        return val



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import warnings
import numpy as np
from collections import defaultdict
from sageopt.coniclifts import utilities as util
from sageopt.coniclifts.constraints.constraint import Constraint
from sageopt.coniclifts.base import ScalarVariable


#
#   Core user-facing functions
#


def compile_problem(objective, constraints):
    if not objective.is_affine():
        raise NotImplementedError()
    # Generate a conic system that is feasible iff the constraints are feasible.
    A, b, K, index_map, var_name_to_locs = compile_constrained_system(constraints)
    # Find all Variable objects in the model (user-defined, or auxilliary).
    all_vars = find_variables_from_constraints(constraints + [objective == 0])
    # Generate the vector for the objective function.
    c, c_offset = compile_linear_expression(objective, all_vars, index_map)
    if c_offset != 0:
        warnings.warn('A constant-term ' + str(c_offset) + ' is being dropped from the objective.')
    c = np.array(c.todense()).flatten().astype(float)
    return c, A, b, K, var_name_to_locs, all_vars


def compile_constrained_system(constraints):
    if any(not isinstance(c, Constraint) for c in constraints):
        raise RuntimeError('compile_constraints( ... ) only accepts iterables of Constraint objects.')
    variables = find_variables_from_constraints(constraints)
    var_gens = np.array([v.generation for v in variables])
    if not np.all(var_gens == var_gens[0]):
        msg1 = '\nThe model contains Variable objects of distinct "generation".\n'
        msg2 = '\nIn between constructing some of these Variables the function \n '
        msg3 = 'coniclifts.clear_variable_indices was called. Remove this function\n'
        msg4 = 'call from your program flow and try again.\n'
        raise RuntimeError(msg1 + msg2 + msg3 + msg4)
    # Categorize constraints (set membership vs elementwise).
    elementwise_constrs, setmem_constrs = [], []
    for c in constraints:
        if c.is_setmem():
            setmem_constrs.append(c)
        elif c.is_elementwise():
            elementwise_constrs.append(c)
        else:
            raise RuntimeError('Unknown argument')
    A, b, K, index_map = conify_constraints(elementwise_constrs, setmem_constrs)
    check_dimensions(A, b, K)
    variables.sort(key=lambda v: v.leading_scalar_variable_id())
    var_indices = []
    for v in variables:
        vids = np.array(v.scalar_variable_ids, dtype=int)
        vi = np.array([index_map[idx] for idx in vids])
        var_indices.append(vi)
    var_name_to_locs = make_variable_map(variables, var_indices)
    return A, b, K, index_map, var_name_to_locs


#
#   Helpers (perform the heavy lifting for core user-facing functions).
#


def conify_constraints(elementwise_constrs, setmem_constrs):
    epigraph_cone_data = epigraph_substitution(elementwise_constrs)
    # Elementwise constraint expressions have now been linearized. Any
    # necessary conic constraints on epigraph variables are in "epigraph_cone_data".
    #
    # The next line simply builds data for linear [in]equality constraints
    # on the linearized versions of the elementwise constraints.
    elementwise_cone_data = [con.conic_form() for con in elementwise_constrs]
    # Elementwise constraints have now been compiled.
    #
    # Now we compile set-membership constraints. Set-membership constraints
    # have customized (possibly sophisticated) compilation functions.
    setmem_cone_data = []
    for con in setmem_constrs:
        setmem_cone_data.extend(con.conic_form())
    # All constraints have been converted to conic form.
    #
    # Now we aggregate this data to facilitate subsequent steps in compilation;
    # we organize constraints as
    #   (1) The original, given elementwise inequalities.
    #   (2) Conic constraints on any auxilliary variables introduced
    #       from epigraph transformations.
    #   (3) The conic versions of the user's "set membership" constraints.
    all_cone_data = elementwise_cone_data + epigraph_cone_data + setmem_cone_data
    matrix_data, bs, K_total = [], [], []
    for A_v, A_r, A_c, b, K in all_cone_data:
        matrix_data.append((A_v, A_r, A_c, b.size))
        bs.append(b)
        K_total.extend(K)
    A, index_map = util.sparse_matrix_data_to_csc(matrix_data)
    b = np.hstack(bs)
    return A, b, K_total, index_map


def epigraph_substitution(elementwise_constrs):
    # Do three things
    #   (1) linearize the Constraint objects by substituting epigraph variables
    #   (2) introduce conic constraints for the epigraph variables
    #   (3) return the conic constraints from (2).
    #
    # This function is not necessary for linear programs, but it shouldn't dramatically
    # slow down LP compilation time either. By calling this function even when all
    # constraints might be linear, we can skip a potentially very expensive curvature check.
    nonlin_atom_to_scalar_exprs = defaultdict(lambda: list())
    for c in elementwise_constrs:
        for se in c.expr.flat:
            for a in se.atoms_to_coeffs:
                if not isinstance(a, ScalarVariable):
                    nonlin_atom_to_scalar_exprs[a].append(se)
        c.epigraph_checked = True
    nl_cone_data = []
    for nl in nonlin_atom_to_scalar_exprs:
        x = nl.epigraph_variable
        A_vals, A_rows, A_cols, b, K = nl.epigraph_conic_form()
        for se in nonlin_atom_to_scalar_exprs[nl]:
            c = se.atoms_to_coeffs[nl]
            del se.atoms_to_coeffs[nl]
            se.atoms_to_coeffs[x] = c
        nl_cone_data.append((A_vals, A_rows, A_cols, b, K))
    return nl_cone_data


def compile_linear_expression(expr, variables, index_map):
    lin_cone_data = (expr == 0).conic_form()
    # The above returns data for the negated expression.
    b = -lin_cone_data[3]
    matrix_data = (lin_cone_data[0], lin_cone_data[1], lin_cone_data[2], b.size)
    A, index_map = util.sparse_matrix_data_to_csc([matrix_data], index_map)
    A = -A
    variables.sort(key=lambda v: v.leading_scalar_variable_id())
    var_indices = []
    for v in variables:
        vids = np.array(v.scalar_variable_ids, dtype=int)
        vi = np.array([index_map[idx] for idx in vids])
        var_indices.append(vi)
    return A, b


def make_variable_map(variables, var_indices):
    """
    :param variables: a list of Variable objects with is_proper=True. These variables
    appear in some vectorized conic system represented by {x : A @ x + b \in K}.

    :param var_indices: a list of 1darrays. The i^th 1darray in this list contains
    the locations of the i^th Variable's entries with respect to the vectorized
    conic system {x : A @ x + b \in K}.

    :return: a dictionary mapping Variable names to ndarrays of indices. These ndarrays
    of indices allow the user to access a Variable's value from the vectorized conic system
    in a very convenient way.

    For example, if "my_var" is the name of a Variable with shape (10, 2, 1, 4), and "x" is
    feasible for the conic system {x : A @ x + b \in K}, then a feasible value for "my_var"
    is the 10-by-2-by-1-by-4 array given by x[user_variable_map['my_var']].

    NOTES:

        This function assumes that every entry of a Variable object is an unadorned ScalarVariable.
        As a result, skew-symmetric Variables or Variables with sparsity patterns are not supported
        by this very important function. Symmetric matrices are supported by this function.

    """
    variable_map = dict()
    for i, v in enumerate(variables):
        temp = np.zeros(v.shape)
        j = 0
        for tup in util.array_index_iterator(v.shape):
            temp[tup] = var_indices[i][j]
            j += 1
        variable_map[v.name] = np.array(temp, dtype=int)
    return variable_map


def find_variables_from_constraints(constraints):
    """
    :param constraints: a list of Constraint objects.
    :return: a list of all coniclifts Variable objects appearing
    in at least one of the Constraints.
    """
    variable_ids = set()
    variables = []
    for c in constraints:
        for v in c.variables():
            if id(v) not in variable_ids:
                variable_ids.add(id(v))
                variables.append(v)
    return variables


def check_dimensions(A, b, K):
    total_K_len = sum([co.len for co in K])
    if total_K_len != A.shape[0]:
        msg = 'K specifies a ' + str(total_K_len) + ' dimensional space, but A has ' + str(A.shape[0]) + ' rows.'
        raise RuntimeError(msg)
    if total_K_len != b.size:
        msg = 'K specifies a ' + str(total_K_len) + ' dimensional space, but b is of length ' + str(b.size) + '.'
        raise RuntimeError(msg)
    pass



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
from sageopt.coniclifts.base import NonlinearScalarAtom, Expression, ScalarExpression, Variable, ScalarVariable
from sageopt.coniclifts.cones import Cone


def weighted_sum_exp(c, x):
    """
    Return a coniclifts ScalarExpression representing the signomial

        sum([ci * e^xi for (ci, xi) in (c, x)])

    :param c: a numpy ndarray of nonnegative numbers.
    :param x: a coniclifts Expression of the same size as c.
    """
    if not isinstance(x, Expression):
        x = Expression(x)
    if not isinstance(c, np.ndarray):
        c = np.array(c)
    if np.any(c < 0):
        raise RuntimeError('Epigraphs of non-constant signomials with negative terms are not supported.')
    if x.size != c.size:
        raise RuntimeError('Incompatible arguments.')
    x = x.ravel()
    c = c.ravel()
    kvs = []
    for i in range(x.size):
        if c[i] != 0:
            kvs.append((Exponential(x[i]), c[i]))
    d = dict(kvs)
    se = ScalarExpression(d, 0, verify=False)
    expr = se.as_expr()
    return expr


class Exponential(NonlinearScalarAtom):

    _EXPONENTIAL_COUNTER_ = 0

    @staticmethod
    def __atom_text__():
        return 'Exponential'

    def __init__(self, x):
        """
        Used to represent the epigraph of "e^x"
        :param x:
        """
        self._args = (self.parse_arg(x),)
        self._id = Exponential._EXPONENTIAL_COUNTER_
        Exponential._EXPONENTIAL_COUNTER_ += 1
        v = Variable(shape=(), name='_exp_epi_[' + str(self.id) + ']_')
        self._epigraph_variable = v[()].scalar_variables()[0]
        pass

    def is_convex(self):
        return True

    def is_concave(self):
        return False

    def epigraph_conic_form(self):
        """
        Refer to coniclifts/standards/cone_standards.txt to see that
        "(x, y) : e^x <= y" is represented as "(x, y, 1) \in K_{exp}".
        :return:
        """
        b = np.zeros(3,)
        K = [Cone('e', 3)]
        A_rows, A_cols, A_vals = [], [], []
        x = self.args[0]
        # first coordinate
        A_rows += (len(x) - 1) * [0]
        A_cols = [var.id for var, co in x[:-1]]
        A_vals = [co for var, co in x[:-1]]
        b[0] = x[-1][1]
        # second coordinate
        A_rows.append(1),
        A_cols.append(self._epigraph_variable.id)
        A_vals.append(1)
        # third coordinate (zeros for A, but included to infer correct dims later on)
        A_rows.append(2)
        A_cols.append(ScalarVariable.curr_variable_count() - 1)
        A_vals.append(0)
        b[2] = 1
        return A_vals, np.array(A_rows), A_cols, b, K

    def value(self):
        x_list = self.args[0]
        d = dict(x_list[:-1])
        x_se = ScalarExpression(d, x_list[-1][1], verify=False)
        x_val = x_se.value()
        val = np.exp(x_val)
        return val



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
from itertools import combinations_with_replacement
from sageopt.symbolic.signomials import Signomial
from sageopt import coniclifts as cl
from sageopt.relaxations import symbolic_correspondences as sym_corr
import warnings


def up_to_q_fold_cons(cons, q):
    cons = [g for g in cons]  # shallow copy
    if len(cons) == 0 or q == 1:
        return cons
    else:
        fold_cons = set()
        for qq in range(q + 1):
            if qq == 0:
                continue
            for comb in combinations_with_replacement(cons, qq):
                temp = np.prod(comb)
                # noinspection PyUnresolvedReferences
                if np.count_nonzero(temp.c) > 1:
                    fold_cons.add(temp)
        fold_cons = list(fold_cons)
    return fold_cons


def valid_posynomial_inequalities(gs):
    conv_gs = []
    for g in gs:
        # g defines a constraint g(x) >= 0
        num_pos = np.count_nonzero(g.c > 0)
        if num_pos >= 2:
            # cannot convexify
            continue
        elif num_pos == 0:
            raise RuntimeError('Attempting to convexify an infeasible signomial inequality constraint.')
        else:
            pos_loc = np.where(g.c > 0)[0][0]
            inverse_term = Signomial({tuple(-g.alpha[pos_loc, :]): 1})
            conv_gs.append(g * inverse_term)
    return conv_gs


def valid_monomial_equations(eqs):
    conv_eqs = []
    for g in eqs:
        # g defines a constraint g(x) == 0.
        if np.count_nonzero(g.c) > 2:
            # cannot convexify
            continue
        pos_loc = np.where(g.c > 0)[0]
        if pos_loc.size == 1:
            pos_loc = pos_loc[0]
            inverse_term = Signomial({tuple(-g.alpha[pos_loc, :]): 1})
            conv_eqs.append(g * inverse_term)
    return conv_eqs


def valid_gp_representable_poly_inequalities(gs):
    gp_rep_polys = []
    for g in gs:
        is_even = len(g.even_locations()) == g.m
        num_pos = np.count_nonzero(g.c > 0)
        if num_pos == 1 and is_even:
            gp_rep_polys.append(g)
        elif num_pos == 0 and is_even:
            if g(np.zeros(g.n)) == 0:
                warnings.warn('The polynomial "g" is only nonnegative at the origin.')
            else:
                raise RuntimeError('Infeasible polynomial inequality.')
    return gp_rep_polys


def valid_gp_representable_poly_eqs(eqs):
    gp_rep_polys = []
    for g in eqs:
        is_even = len(g.even_locations()) == g.m
        if is_even and np.count_nonzero(g.c != 0) == 2 and np.count_nonzero(g.c > 0) == 1:
            gp_rep_polys.append(g)
    return gp_rep_polys


def unconstrained_cut(v, alpha, ell=1, AbK=None, solver='MOSEK'):
    # v is supposed to belong to convco{ exp(alpha @ x) : x in Rn }, but instead
    # it only belongs to some outer-approximation. We need to find a separating hyperplane
    # (in terms of a signomial coefficient vector) that rules out v.
    #
    # Use this function to improve a bound in any problem (constrained or unconstrained)
    # where the gap between SAGE and nonnegativity is viewed as the likely source of error.
    #
    # When using this function with a constrained problem, "v" and "alpha" should be associated
    # with the constraint "Lagrangian is SAGE".
    from sageopt.relaxations.sage_sigs import primal_sage_cone
    from sageopt.coniclifts.operators.norms import vector2norm
    c_x = cl.Variable(shape=(alpha.shape[0],))
    s_x = Signomial(alpha, c_x)
    modulator = Signomial(alpha, np.ones(shape=(alpha.shape[0],))) ** ell
    modded_s_x = s_x * modulator
    constraints = [primal_sage_cone(modded_s_x, name='sig_is_nonneg', AbK=AbK),
                   vector2norm(c_x) <= np.linalg.norm(v)]
    objective_expr = c_x @ v
    prob = cl.Problem(cl.MIN, objective_expr, constraints)
    prob.solve(solver=solver, verbose=False)
    cl.clear_variable_indices()
    if prob.status == 'solved' and prob.value < 0:
        c_val = c_x.value()
        cut_sig = Signomial(alpha, c_val)
        return cut_sig
    else:
        return None


def constrained_cut(v, alpha, gts, eqs, ell=1, solver='MOSEK'):
    #
    # Would be nice to infer a valid inequality that strengthens the Lagrange dual
    # (not just the SAGE relaxation of the Lagrange dual). Right now this function
    # only strengthens a given SAGE relaxation, and makes it at-most-as-strong-as
    # the Lagrange dual.
    #
    #
    from sageopt.relaxations.sage_sigs import primal_sage_cone
    from sageopt.coniclifts.operators.norms import vector2norm
    c_x0 = cl.Variable(shape=(alpha.shape[0],))
    s_x0 = Signomial(alpha, c_x0)
    modulator = Signomial(alpha, np.ones(shape=(alpha.shape[0],))) ** ell
    modded_s_x0 = s_x0 * modulator
    constraints = [primal_sage_cone(modded_s_x0, name='sig_is_nonneg', AbK=None)]
    c_x = c_x0
    if len(gts) > 0:
        lambda_vars = cl.Variable(shape=(len(gts),))
        constraints.append(lambda_vars >= 0)
        G = np.hstack([sym_corr.relative_coeff_vector(g, alpha).reshape((-1, 1)) for g in gts])
        c_g = G @ lambda_vars
        c_x = c_x + c_g
    if len(eqs) > 0:
        mu_vars = cl.Variable(shape=(len(eqs),))
        H = np.hstack([sym_corr.relative_coeff_vector(h, alpha).reshape((-1, 1)) for h in eqs])
        c_h = H @ mu_vars
        c_x = c_x + c_h
    constraints.append(vector2norm(c_x) <= np.linalg.norm(v))
    objective_expr = c_x @ v
    prob = cl.Problem(cl.MIN, objective_expr, constraints)
    prob.solve(solver=solver, verbose=False)
    cl.clear_variable_indices()
    if prob.status == 'solved' and prob.value < 0:
        c_val = c_x.value()
        cut_sig = Signomial(alpha, c_val)
        return cut_sig
    else:
        return None
